/* this is the implementation of FELL -- Front End Little Lisp

   Copyright 2001/2002.  Rodney A. Brooks
   This code is only intended to be used in class 6.836 at MIT,
   in the spring term 2002.

   it is written in a very benign version of ANSI C, and should be compilable
   just about anywhere.  it implements a subset of Common Lisp.  there
   is no compiler, but there is now a garbage collector.

   it is intended to be called with initlisp(<number>, <string>) where <number>
   is how many K should be included in the heap.  it expects to find a file
   <string> in the default directory.  everything in that file gets
   evaluated at initialization time and completes the definition of FELL.
   <string> can also be NULL, in which case no startup file is loaded

   then fell_reploop(<string>) is called to start the read eval print loop.
   <string> is another file which can get evaluated at that time.

   it expects to have an outcall to background(<status>), which lets something
   else run in the background. <status> tells the background process whether
   FELL is really busy or not, so that it can decide how hard to run.

   there is a mechanism for putting c procedures into the lisp name space,
   so that they can be called from here.

   in the reploop type *symbols* to see a hash vector of all the symbols.
   *specials* also gives some clues as to what is here.

   the reader handles # and backquote.  macros are available.

*/

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <math.h>
#include "fell.h"

/*****************************************************************************
******************************************************************************
******************************************************************************

*** declarations of all the types and data structures

******************************************************************************
******************************************************************************
*****************************************************************************/

typedef enum {type_null, type_cons, type_fixnum, type_flonum,
	      type_string, type_symbol, type_vector, type_structure,
	      type_procedure, type_marker, type_cpointer, type_filler,
	      type_last
} lisptype;
			
#define max_ltypes	100

typedef enum {class_object, class_cstring} storageclass;

typedef enum {mark_unbound_value, mark_undef_proc} markerclass;
		
typedef int *object; 

typedef struct
	{object car;
	 object cdr;}
	pair;
	
typedef struct
	{int value;}
	fixnum;
	
typedef struct
	{double value;}
	flonum;

typedef struct
	{object name;
	 object value;
	 object procedure;}
	symbol;
	
typedef struct
	{int length;
	 char chars[1];}
	string;
	
typedef struct
	{int length;
	 int *arr[1];}
	vector;
	
typedef struct
	{int numslots;
	 object descriptor;
	 int *slots[1];}
	 structure;
	 
typedef struct
	{int info;
	 object name;
	 void *proc;
	 object arglist;
	 object env;}
	 procedure;
	
typedef struct
	{int mark;}
	marker;
	
typedef struct
	{void *point;}
	cpointer;

typedef enum {struct_type, struct_numslots, struct_slotnames,
			  struct_slotdefaults}
	structslots;
	
/* accessors for parts of lisp objects */

#define headeritems		1

#define gettype(obj)	*obj
#define getltypestring(obj)	((char *)*(obj+1))
#define pair(obj)		((pair *)(obj+headeritems))
#define fixnum(obj)		((fixnum *)(obj+headeritems))
#define flonum(obj)		((flonum *)(obj+headeritems))
#define symbol(obj)		((symbol *)(obj+headeritems))
#define string(obj)		((string *)(obj+headeritems))
#define marker(obj)		((marker *)(obj+headeritems))
#define vector(obj)		((vector *)(obj+headeritems))
#define structure(obj)	((structure *)(obj+headeritems))
#define procedure(obj)	((procedure *)(obj+headeritems))
#define cpointer(obj)	((cpointer *)(obj+headeritems))

#define car(obj)		pair(obj)->car
#define cdr(obj)		pair(obj)->cdr
#define first(obj)		pair(obj)->car
#define second(obj)		pair(pair(obj)->cdr)->car
#define third(obj)		pair(pair(pair(obj)->cdr)->cdr)->car
#define rest(obj)		pair(obj)->cdr
#define structref(obj,slot)		structure(obj)->slots[slot]
#define vref(obj,index)	vector(obj)->arr[index]

#define list1(arg)		cons(arg,nilobj)
#define list2(a1,a2)	cons(a1,cons(a2,nilobj))
#define list3(a1,a2,a3)	cons(a1,cons(a2,cons(a3,nilobj)))

#define nullp(arg)		(arg == nilobj)
#define pairp(arg)		(type_cons == gettype(arg))
#define notpairp(arg)	(type_cons != gettype(arg))
#define symbolp(arg)	(type_symbol == gettype(arg))
#define stringp(arg)	(type_string == gettype(arg))
#define fixnump(arg)	(type_fixnum == gettype(arg))
#define flonump(arg)	(type_flonum == gettype(arg))
#define vectorp(arg)	(type_vector == gettype(arg))
#define structurep(arg)	(type_structure == gettype(arg))
#define procedurep(arg)	(type_procedure == gettype(arg))
#define cpointerp(arg)	(type_cpointer == gettype(arg))
#define boundp(arg)		(unbound_value!=symbol(arg)->value)

#define hashlooker(procname)	(object(*)(object))procname

#define obtablesize		4096
#define structtablesize		256
#define specialstablesize	512
#define constantstablesize	512

#define proc_primop		0x10000
#define proc_lambda		0x20000
#define proc_macro		0x40000
#define proc_cproc		0x80000

#define maxargs	255

#define make_procinfo(type,arginfo)	(type|arginfo)
#define make_arginfo(min,max)		(min|(max<<8))
#define minargcount(proc)		(0xff&(procedure(proc)->info))
#define maxargcount(proc)		(0xff&((procedure(proc)->info)>>8))
#define primopp(proc)	(proc_primop&(procedure(proc)->info))
#define lambdap(proc)	(proc_lambda&(procedure(proc)->info))
#define macroprocp(proc)	(proc_macro&(procedure(proc)->info))
#define procp(proc)	((proc_primop|proc_lambda|proc_cproc)&(procedure(proc)->info))
#define cprocp(proc)	(proc_cproc&(procedure(proc)->info))

#define true (1)
#define false (0)
typedef int boolean;

/* things that get set up for the world during initialization */

static object nilobj;
static object zerobj;
static object unbound_value;
static object undef_proc;
static object obtable;
static object obtablesym;
static object structtable;
static object structtablesym;
static object structuresym;
static object specials;
static object specialssym;
static object constants;
static object constantssym;
static object ltypes;
static object ltypessym;
static object printbase;
static object readbase;
static object features;
static object andsymbol;
static object orsymbol;
static object notsymbol;
static object nullsymbol;
static object quote;
static object function;
static object tsymbol;
static object printlevel;
static object printlength;
static object printarray;
static object printstructure;
static object structurestructure;
static object stringkeyword;
static object minus;
static object plusplusplus;
static object plusplus;
static object plus;
static object starstarstar;
static object starstar;
static object star;
static object ifsymbol;
static object prognsymbol;
static object whensymbol;
static object unlesssymbol;
static object blocksymbol;
static object returnsymbol;
static object returnfromsymbol;
static object inputkey;
static object outputkey;
static object appendkey;
static object supersedekey;
static object directkey;
static object ifexistskey;
static object filekey;
static object defunsymbol;
static object defmacrosymbol;
static object defconstantsymbol;
static object defvarsymbol;
static object defparametersymbol;
static object andrestsymbol;
static object andoptsymbol;
static object andkeysymbol;
static object framesymbol;
static object catchsymbol;
static object throwsymbol;
static object unwindsymbol;
static object setqsymbol;
static object letsymbol;
static object progvsymbol;
static object debugprintlevel;
static object debugprintlength;
static object debugprintbase;
static object inspectprintlevel;
static object inspectprintlength;
static object inspectprintbase;
static object initelemkey;
static object tagbodysymbol;
static object gosymbol;
static object dotimessymbol;
static object dolistsymbol;
static object gensymcounter;
static object lambda;
static object startkey;
static object endkey;
static object evalsymbol;

static object l__splice;
static object l__dot;
static object l__comma;
static object l__clobber;
static object l__nconc;
static object l__append;
static object l__list;
static object l__liststar;


/* io interface types */

typedef struct
	{char *filename;
	 int current;
	 int limit;
	 int lastchar;
	 boolean openstatus;
	 void *prmeth;
	 FILE *openfile;}
	 streamstruct;

typedef streamstruct *lstream;

/* used to implement catch/throw in the reader */

typedef struct
	{jmp_buf env;
	 boolean readinternal;
	 int bqlevel;
	 void *rparen;
	 void *dot;
	 void *doterror;
	 void *eof;
	 int free;
	 }
	 rcatchpoint;

/* used by the printer */

typedef struct
	{int count;
	 char *string;
	 int size;
	 int column;
	 void *stream;
	 }
	 *pmethod;
	 
static pmethod tprintmethod;

/* general catch/throw/unwind-protect used by eval */

typedef struct
	{jmp_buf env;
	 int free;
	 }
	 catchpoint;


/* for the memory manager */

typedef struct memspacedesc *memspace;

struct memspacedesc {
  char *lowest;
  char *freepoint;
  int size;
  int freemem;};


/* things for the eval stack */

#define stacksize 4096
#define stackbound (stacksize-100)
#define regstacksize 8192
#define regstackbound (regstacksize-200)

static object stack[stacksize];
static object regstack[regstacksize];
static int sp, regsp;
static object prompt;

typedef struct savstack {
  int spsave;
  int regspsave;} sstack;

#define makeframe(pname,env)	list2(framesymbol,cons(pname,env))
#define frameproc(frame)	first(second(frame))
#define lexenv(frame)		rest(second(frame))
#define pushframe(var,val,frame) cdr(cdr(frame))=cons(cons(var,val),cdr(cdr(frame)))
#define pushletbind(var,val,letb) cdr(letb)=cons(cons(var,val),cdr(letb))
#define pushtag(tag,tagb)	cdr(tagb)=cons(cons(tag,tagb),cdr(tagb))
#define pushstackitem(item,cnt)	stack[++sp]=cons(item,cnt)
#define pushstack(item)		pushstackitem(item,stack[sp])
#define pushstackfresh(item)	pushstackitem(item,nilobj)
#define savestack(place)	(place.spsave=sp,place.regspsave=regsp)
#define restorestack(place)	(sp=place.spsave,regsp=place.regspsave)
#define savereg(item)		regstack[++regsp]=item
#define restorereg(reg)		reg = regstack[regsp--]
#define getbackreg(index,reg)	reg = regstack[regsp-index]
#define referreg(index)		regstack[regsp-index]

/*****************************************************************************
******************************************************************************
******************************************************************************

*** the prototypes for all the procedures in this file

******************************************************************************
******************************************************************************
*****************************************************************************/

static memspace make_memspace (int);
static void init_procedures ();
static object room (object);
static void room_space (char *, memspace);
static void quit ();
static int *rawalloc (int);
static int *crawalloc (int);
static object lalloc (lisptype, storageclass);
static object lallocvar (lisptype, storageclass, int);
static object address_of_proc (object);
static object display_memory_proc (object, object);
static object objectify (object);
static object make_cpointer (void *point);
static object make_fixnum (int num);
static object make_flonum (double num);
static object make_string (char *str);
static object make_symbol (char *name, object value, object proc);
static object cons (object car, object cdr);
static object make_procedure (int info);
static object make_vector (int length);
static object make_structure (object descriptor);
static object struct_desc_name (object descriptor);
static object make_structure_structure ();
static void register_structure (object descriptor);
static int get_hashindex (object vec, char *str, object (*lookup)(object));
static int compute_hash (char *str);
static boolean different_strings (char *str1, object entry);
static object symbol_name (object symb);
static object intern (char *str);
static void declare_special(object sym);
static object is_special_p (object sym);
static void declare_constant (object sym);
static object is_constant_p (object sym);

static object gcprint ();
static object gc ();
static void gc_sweep (memspace, memspace);
static object gc_forward (object, memspace);

static object print (object obj, object stream);
static object prin1 (object obj, object stream);
static object princ (object obj, object stream);
static object terpri (object stream);
static object printproc (object obj, object restargs);
static object prin1proc (object obj, object restargs);
static object princproc (object obj, object restargs);
static object terpriproc (object restargs);
static object format (object stream, object strng, object restargs);
static void prin1_float_size (double number,
			      pmethod method, int size, int places);
static void outtab(boolean flag, int number, pmethod method);
static void princ_aux (object obj, pmethod method);
static void prin1_aux (object obj, pmethod method);
static void princ_string_c (char *str, pmethod method);
static void prin1_float_c (double f, pmethod method);
static void prin1_fixnum_main_c (int num,
				 pmethod method,
				 int trail, char trailchar, int base);
static void prin1_fixnum_c (int num,
			    pmethod method,
			    int trail, char trailchar, int base);
static void initialize_printer();
static pmethod new_pmethod();
static pmethod get_print_method (object stream);
static void outchar (char ch, pmethod method);
static void force_output (pmethod method);

static rcatchpoint *get_freercatchpoint ();
static void restore_rcatchpoint (rcatchpoint *catchp);
static void init_reader ();
static object fellread (object restargs);
static object readproc (object stream, object eof_error_p, object eof_value);
static lstream get_read_method (object stream);
static object read_from_string (object string,
				object eof_error_p,
				object eof_value, int start, int end);
static object read_aux ();
static object read_string ();
static object read_from_string_proc (object str, object restargs);
static object read_pound_sign();
static object read_aux_with_base (int base);
static object read_decide_on_conditional (object testexp, object actualexp);
static boolean read_conditional_truep (object testexp);
static object read_is_number_p(char *str);
static object read_backquote();
static object read_comma();
static object read_process_backquote (object form);
static object read_bracket (object form);
static object read_maptree_simplify (object x);	    
static object read_simplify (object x);
static object read_attach_append (object op, object item, object result);
static object read_attach_conses (object items, object result);	
static object mapsecond (object lst);
static int nextchar (int tokenp);
static void backchar (char ch);
static object fellopen (object filename, object restvars);
static int felllength (object lst);
static void checkkeywordargs (object args, int arglength, object validkeys);
static object fellclose (object stream);
static void read_error (char *mess, char ch);
static void read_error_long ();
static void read_lerror (char ch, int attr);
static void read_general_error (char *str);
static void unexpected_end_error (char *str);
static object end_of_file_error (lstream stream,
				 object eof_error_p, object eof_return);

static object memq (object item, object lst);
static object assq (object item, object lst);
static object append (object u, object v);
static object nreverse (object lst);
static object reverse (object lst);
static object nconc (object lst1, object lst2);
static object nthproc (object, object);
static object setproc (object, object);
static object atomproc (object arg);
static object conspproc (object arg);
static object nullproc (object arg);
static object symbolpproc (object arg);
static object vectorpproc (object arg);
static object stringpproc (object arg);
static object procedurepproc (object arg);
static object structurepproc (object arg);
static object fixnumpproc (object arg);
static object flonumpproc (object arg);
static object numberpproc (object arg);
static object eqproc (object arg1, object arg2);
static object numeqproc (object item1, object item2);
static object numneproc (object item1, object item2);
static object numltproc (object item1, object item2);
static object numleproc (object item1, object item2);
static object numgtproc (object item1, object item2);
static object numgeproc (object item1, object item2);
static boolean bequal (object item1, object item2);
static object equalproc (object item1, object item2);
static object carproc (object x);
static object cdrproc (object x);
static object secondproc (object x);
static object thirdproc (object x);
static object listproc (object restarg);
static object liststarproc (object fst, object restargs);
static object fell_boundp (object sym);
static object keywordp (object sym);
static object adder (object restargs);
static object subber (object firstarg, object restargs);
static object timeser (object restargs);
static object divider (object firstarg, object restargs);
static object sqrtproc (object);
static object lshproc (object, object);
static object floatproc (object);
static object logander (object restargs);
static object logiorer (object restargs);
static object logxorer (object restargs);
static object lognoter (object arg);
static object svrefproc (object vector, object index);
static object svsetproc (object vector, object index, object val);
static void check_vector_args (object vec, object index);
static object mvecproc (object size, object restargs);
static object lengthproc (object obj);
static object gensymproc ();
static object mapcarproc (object proc, object lst1, object restargs);
static object check_ltype_proc (object obj, object str);
static void init_error_handlers ();
static void error0 (char *errorstr);
static void error1 (char *errorstr, object item);
static void error2 (char *errorstr, object item1, object item2);
static object errorproc (object string, object restargs);
static void warn1 (char *warnstr, object item);
static void warn2 (char *warnstr, object item, object item2);
static object warnproc (object string, object restargs);
static void error_needallnumbers (object restargs);
static void error_needallfixnums (object restargs);
static void check_listnil (object lst);
static void check_cons (object conshuh);
static void check_consnil (object, char *);
static void check_number (object num);
static void check_fixnum (object);
static void error_recover (object str, object lst);
static void enter_debugger ();
static boolean debug_valid_p (object item, boolean nilok, int minnum);
static void exit_debugger ();
static void display_frame (int index);

static catchpoint *get_freecatchpoint ();
static void restore_catchpoint (catchpoint *catchp);
static object loadfile (object filename);
static object eval(object form);
static object applyproc (object proced, object arglist);
static object applyspread (object proced, object restargs);
static object apply_cproc (object proced, object restargs);
static void *process_carg(int typ, object arg, double *stash);
static object defineproc (object, boolean, int, object, object);
static object eval_form(object form);
static object eval_form_list (object forms);
static object eval_list (object forms);
static void eval_return (object blockname, object form);
static void simthrow (object tag, object value);
static object parse_tagbody (object forms);
static void eval_tagbody (object forms, object catcher);
static void check_specialform (int len, int lo, int hi, object form);
static void check_doform (object subf);
static void check_arglist (object arglist);
static object bindargs (object formals,
			object actuals, object proced, object arglist);
static object searchvarstack (object, object);
static int check_letbindings (object pairforms);
static object eval_letbindings (object pairforms);
static object lookup_gotag (object tag);
static object macroexpand1 (object);
static object inspect (object);
static void exit_inspector ();
static void inspect_item (object);


/*****************************************************************************
******************************************************************************
******************************************************************************

*** the memory management system and initialization of FELL

******************************************************************************
******************************************************************************
*****************************************************************************/

static memspace make_memspace (int spacek) {
  int spacesize;
  memspace space;

  space = (memspace)malloc(sizeof(struct memspacedesc));
  spacesize = spacek<<10;

  space->freepoint
    = space->lowest
    = (char *)((~3)&(3+((int)malloc(spacesize+3))));

  if (space->lowest == 0)
    {printf("\nNo memory!\n");
    exit(1);}

  space->freemem = space->size = spacesize;

  return space;}


static memspace staticspace, tospace, fromspace, cspace;

static int gc_count;

static int headersize;
#define hashedfixnumsk 2
#define fixnumoffset (hashedfixnumsk<<10)

static object fixarray[2*fixnumoffset];


/* initialize the lisp heap and the environment */

extern void initlisp (int heapk, void *filename) {
  int i;
  object fix;

  headersize = sizeof(int);
  staticspace = make_memspace(64);
  cspace = make_memspace(256);
  gc_count = 0;

  /* set up the hashed fixnums */
  tospace = make_memspace(2*hashedfixnumsk*(headersize+sizeof(fixnum)));
  for(i=-fixnumoffset;i<fixnumoffset;i++) {
    fixarray[i+fixnumoffset] = fix = lalloc(type_fixnum, class_object);
    fixnum(fix)->value = i;}
  zerobj = fixarray[fixnumoffset];
  
  /* set up the initial environment -- do it in static space */
  tospace = staticspace;
  nilobj = lalloc(type_null, class_object);
  pair(nilobj)->car = nilobj;
  pair(nilobj)->cdr = nilobj;
  
  unbound_value = lalloc(type_marker, class_object);
  marker(unbound_value)->mark = mark_unbound_value;
  
  undef_proc = lalloc(type_marker, class_object);
  marker(undef_proc)->mark = mark_undef_proc;
  
  obtable = make_vector(obtablesize);
  obtablesym = make_symbol("*symbols*", obtable, undef_proc);
  structtable = make_vector(structtablesize);
  structtablesym = make_symbol("*structs*", structtable, undef_proc);
  specials = make_vector(specialstablesize);
  specialssym = make_symbol("*specials*", specials, undef_proc);
  constants = make_vector(constantstablesize);
  constantssym = make_symbol("*constants*", constants, undef_proc);
  ltypes = make_vector(max_ltypes-type_last);
  ltypessym = make_symbol("*ltypes*", ltypes, undef_proc);
  
  printbase = make_symbol("*print-base*", make_fixnum(10), undef_proc);
  readbase = make_symbol("*read-base*", make_fixnum(10), undef_proc);
  features = make_symbol("*features*", nilobj, undef_proc);
  andsymbol = make_symbol("and", unbound_value, undef_proc);
  orsymbol = make_symbol("or", unbound_value, undef_proc);
  notsymbol = make_symbol("not", unbound_value, undef_proc);
  nullsymbol = make_symbol("null", unbound_value, undef_proc);
  quote = make_symbol("quote", unbound_value, undef_proc);
  function = make_symbol("function", unbound_value, undef_proc);
  tsymbol = make_symbol("t", unbound_value, undef_proc);
  symbol(tsymbol)->value = tsymbol;
  printlevel = make_symbol("*print-level*", nilobj, undef_proc);
  printlength = make_symbol("*print-length*", nilobj, undef_proc);
  printarray = make_symbol("*print-array*", tsymbol, undef_proc);
  printstructure = make_symbol("*print-structure*", tsymbol, undef_proc);
  structurestructure = make_structure_structure();
  stringkeyword = intern(":string");
  minus = make_symbol("-", nilobj, undef_proc);
  plusplusplus = make_symbol("+++", nilobj, undef_proc);
  plusplus = make_symbol("++", nilobj, undef_proc);
  plus = make_symbol("+", nilobj, undef_proc);
  starstarstar = make_symbol("***", nilobj, undef_proc);
  starstar = make_symbol("**", nilobj, undef_proc);
  star = make_symbol("*", nilobj, undef_proc);
  ifsymbol = intern("if");
  prognsymbol = intern("progn");
  whensymbol = intern("when");
  unlesssymbol = intern("unless");
  blocksymbol = intern("block");
  returnsymbol = intern("return");
  returnfromsymbol = intern("return-from");
  inputkey = intern(":input");
  outputkey = intern(":output");
  appendkey = intern(":append");
  supersedekey = intern(":supersede");
  directkey = intern(":direction");
  ifexistskey = intern(":if-exists");
  filekey = intern(":file");
  defunsymbol = intern("defun");
  defmacrosymbol = intern("defmacro");
  defconstantsymbol = intern("defconstant");
  defvarsymbol = intern("defvar");
  defparametersymbol = intern("defparameter");
  andrestsymbol = intern("&rest");
  andoptsymbol = intern("&optional");
  andkeysymbol = intern("&key");
  framesymbol = intern(":frame");
  catchsymbol = intern("catch");
  throwsymbol = intern("throw");
  unwindsymbol = intern("unwind-protect");
  setqsymbol = intern("setq");
  letsymbol = intern("let");
  progvsymbol = intern("progv");
  debugprintlevel = make_symbol("*debug-print-level*",
				make_fixnum(3), undef_proc);
  debugprintlength = make_symbol("*debug-print-length*",
				 make_fixnum(6), undef_proc);
  debugprintbase = make_symbol("*debug-print-base*",
			       make_fixnum(10), undef_proc);
  inspectprintlevel = make_symbol("*inspect-print-level*",
				  make_fixnum(3), undef_proc);
  inspectprintlength = make_symbol("*inspect-print-length*",
				   make_fixnum(6), undef_proc);
  inspectprintbase = make_symbol("*inspect-print-base*",
				 make_fixnum(10), undef_proc);
  initelemkey = intern(":initial-element");
  tagbodysymbol = intern("tagbody");
  gosymbol = intern("go");
  dotimessymbol = intern("dotimes");
  dolistsymbol = intern("dolist");
  gensymcounter = make_symbol("*gensym*", zerobj, undef_proc);
  lambda = intern("lambda");
  startkey = intern(":start");
  endkey = intern(":end");
  evalsymbol = intern("eval");
  prompt = make_string("~&#? ");
  
  declare_special(obtablesym);
  declare_special(structtablesym);
  declare_special(specialssym);
  declare_special(constantssym);
  declare_special(printbase);
  declare_special(readbase);
  declare_special(features);
  declare_special(printlevel);
  declare_special(printlength);
  declare_special(printarray);
  declare_special(printstructure);
  declare_special(minus);
  declare_special(plusplusplus);
  declare_special(plusplus);
  declare_special(plus);
  declare_special(starstarstar);
  declare_special(starstar);
  declare_special(star);
  declare_special(debugprintlevel);
  declare_special(debugprintlength);
  declare_special(debugprintbase);
  declare_special(inspectprintlevel);
  declare_special(inspectprintlength);
  declare_special(inspectprintbase);
  declare_special(gensymcounter);
  
  declare_constant(tsymbol);
  
  l__splice = make_symbol("%%splice", unbound_value, undef_proc);
  l__dot = make_symbol("%%dot", unbound_value, undef_proc);
  l__comma = make_symbol("%%comma", unbound_value, undef_proc);
  l__clobber = make_symbol("%%clobber", unbound_value, undef_proc);
  l__nconc = make_symbol("%%nconc", unbound_value, undef_proc);
  l__append = make_symbol("%%append", unbound_value, undef_proc);
  l__list = make_symbol("%%list", unbound_value, undef_proc);
  l__liststar = make_symbol("%%list*", unbound_value, undef_proc);
  
  sp = 0;
  regsp = 0;
  stack[sp] = nilobj;
  regstack[sp] = nilobj;

  initialize_printer();
  init_error_handlers();
  init_reader();
  init_procedures();
  printf("Started up Front End Little Lisp (FELL)...");
  
  tprintmethod->column = 0;
  if (NULL!=filename)
    loadfile(make_string((char *)filename));
  else
    printf("\n");
  tospace = make_memspace(heapk);		/* set up regular memory */
  fromspace = make_memspace(heapk);
  return;}

struct procfill {
  char *name;
  void *proc;
  int info;
} filltab[] = {
  "room", room, make_arginfo(0,1),
  "gcprint", gcprint, make_arginfo(1,1),
  "gc", gc, make_arginfo(0,0),
  "quit", quit, make_arginfo(0,0),
  "inspect", inspect, make_arginfo(1,1),
  "nreverse", nreverse, make_arginfo(1,1),
  "reverse", reverse, make_arginfo(1,1),
  "boundp", fell_boundp, make_arginfo(1,1),
  "keywordp", keywordp, make_arginfo(1,1),
  "memq", memq, make_arginfo(2,2),
  "assq", assq, make_arginfo(2,2),
  "append", append, make_arginfo(2,2),
  "nconc", nconc, make_arginfo(2,2),
  "nth", nthproc, make_arginfo(2,2),
  "set", setproc, make_arginfo(2,2),
  "+", adder, make_arginfo(0,maxargs),
  "-", subber, make_arginfo(1,maxargs),
  "*", timeser, make_arginfo(0,maxargs),
  "/", divider, make_arginfo(1,maxargs),
  "sqrt", sqrtproc, make_arginfo(1,1),
  "lsh", lshproc, make_arginfo(2,2),
  "float", floatproc, make_arginfo(1,1),
  "logand", logander, make_arginfo(0,maxargs),
  "logior", logiorer, make_arginfo(0,maxargs),
  "logxor", logxorer, make_arginfo(0,maxargs),
  "lognot", lognoter, make_arginfo(1,1),
  "atom", atomproc, make_arginfo(1,1),
  "consp", conspproc, make_arginfo(1,1),
  "null", nullproc, make_arginfo(1,1),
  "not", nullproc, make_arginfo(1,1),
  "symbolp", symbolpproc, make_arginfo(1,1),
  "vectorp", vectorpproc, make_arginfo(1,1),
  "stringp", stringpproc, make_arginfo(1,1),
  "procedurep", procedurepproc, make_arginfo(1,1),
  "structurep", structurepproc, make_arginfo(1,1),
  "fixnump", fixnumpproc, make_arginfo(1,1),
  "flonump", flonumpproc, make_arginfo(1,1),
  "numberp", numberpproc, make_arginfo(1,1),
  "eq", eqproc, make_arginfo(2,2),
  "=", numeqproc, make_arginfo(2,2),
  "/=", numneproc, make_arginfo(2,2),
  "<", numltproc, make_arginfo(2,2),
  "<=", numleproc, make_arginfo(2,2),
  ">", numgtproc, make_arginfo(2,2),
  ">=", numgeproc, make_arginfo(2,2),
  "car", carproc, make_arginfo(1,1),
  "cdr", cdrproc, make_arginfo(1,1),
  "first", carproc, make_arginfo(1,1),
  "second", secondproc, make_arginfo(1,1),
  "third", thirdproc, make_arginfo(1,1),
  "rest", cdrproc, make_arginfo(1,1),
  "cons", cons, make_arginfo(2,2),
  "list", listproc, make_arginfo(0,maxargs),
  "list*", liststarproc, make_arginfo(1,maxargs),
  "%%append", append, make_arginfo(2,2),
  "%%nconc", nconc, make_arginfo(2,2),
  "%%list", listproc, make_arginfo(0,maxargs),
  "%%list*", liststarproc, make_arginfo(1,maxargs),
  "open", fellopen, make_arginfo(1,5),
  "close", fellclose, make_arginfo(1,1),
  "read", fellread, make_arginfo(0,3),
  "load", loadfile, make_arginfo(1,1),
  "format", format, make_arginfo(2,maxargs),
  "equal", equalproc, make_arginfo(2,2),
  "print", printproc, make_arginfo(1,2),
  "prin1", prin1proc, make_arginfo(1,2),
  "princ", princproc, make_arginfo(1,2),
  "terpri", terpriproc, make_arginfo(0,1),
  "error", errorproc, make_arginfo(1,maxargs),
  "warn", warnproc, make_arginfo(1,maxargs),
  "svref", svrefproc, make_arginfo(2,2),
  "svset", svsetproc, make_arginfo(3,3),
  "make-vector", mvecproc, make_arginfo(1,3),
  "length", lengthproc, make_arginfo(1,1),
  "gensym", gensymproc, make_arginfo(0,0),
  "mapcar", mapcarproc, make_arginfo(2,maxargs),
  "eval", eval_form, make_arginfo(1,1),
  "funcall", applyproc, make_arginfo(1,maxargs),
  "apply", applyspread, make_arginfo(1,maxargs),
  "read-from-string", read_from_string_proc, make_arginfo(1,7),
  "address-of", address_of_proc, make_arginfo(1,1),
  "display-memory", display_memory_proc, make_arginfo(2,2),
  "objectify", objectify, make_arginfo(1,1),
  "check-ltype", check_ltype_proc, make_arginfo(2,2),
  "macroexpand-1", macroexpand1, make_arginfo(1, 1),
  "", NULL, 0};


static void init_procedures () {
  int i;
  object sym, proc;
	
  for (i=0;NULL!=filltab[i].proc;i++)
    {sym = intern(filltab[i].name);
    proc = make_procedure(make_procinfo(proc_primop,filltab[i].info));
    symbol(sym)->procedure = proc;
    procedure(proc)->name = sym;
    procedure(proc)->proc = filltab[i].proc;}
  register_cprocedures();
  return;}


static object room (object restarg) {
  printf("There are %d free bytes, and there have been %d GCs.\n",
	 tospace->freemem,
	 gc_count);
  if (nilobj!=restarg) {
    room_space("Static", staticspace);
    room_space("To    ", tospace);
    printf("From  : %8.8x\n", (int)fromspace->lowest);
    room_space("Cspace", cspace);}
  return tsymbol;}
	

static void room_space (char *name, memspace sp) {
  printf("%s: %8.8x  Size: %8d  Used: %8d  Free: %8d\n",
	 name,
	 (int)sp->lowest,
	 sp->size,
	 sp->size-sp->freemem,
	 sp->freemem);
  return;}


static object warnstring;

static void quit () {
  simthrow(warnstring,nilobj);}	


/* allocate a chunk of storage in the heap */

static int *rawalloc (int size) {
  boolean gcflag;
  int *ret;
	
  gcflag = false;
  size = (~3)&(size+3);
 loop:
  if (tospace->freemem<size) {
    if (gcflag) {
      printf("\nGC did not help!!!  Giving up.\n\n");
      exit(-1);}
    gcflag = true;
    gc();
    goto loop;}
  tospace->freemem-=size;
  ret = (int *) tospace->freepoint;
  tospace->freepoint += size;
  return ret;}


/* allocate a chunk of storage in the C object space */

static int *crawalloc (int size) {
  int *ret;
  size = (~3)&(size+3);
  if (0>(cspace->freemem-=size))
    {printf("\nExhausted free C memory!! %d\n\n", size);
    exit(-1);}
  ret = (int *) cspace->freepoint;
  cspace->freepoint += size;
  return ret;}


static object lalloc (lisptype type, storageclass sclass) {
  int *ret;
  int size;
	
  if (sclass != class_object)
    printf("Can not handle storage class: %d", sclass);
	  
  switch(type) {
  case type_null: case type_cons:
    size = sizeof(pair);
    break;
  case type_fixnum:
    size = sizeof(fixnum);
    break;
  case type_flonum:
    size = 4+sizeof(flonum);
    break;
  case type_symbol:
    size = sizeof(symbol);
    break;
  case type_procedure:
    size = sizeof(procedure);
    break;
  case type_marker:
    size = sizeof(marker);
    break;
  case type_cpointer:
    size = sizeof(cpointer);
    break;
  default:
    printf("Can not allocate type: %d\n", type);
    break;}
	
  ret = rawalloc(size+headersize);
  if (type==type_flonum) {
    /* for 8 byte alignment machines */
    if (0==(4&(int)ret))
      *ret++ = type_filler;
    else
      *(3+ret) = type_filler;}

  *ret = (int) type;
  return (object) ret;}


static object lallocvar (lisptype type, storageclass sclass, int elements) {
  int *ret;
  int size, multiplier;
	
  if (sclass != class_object)
    printf("Can not handle storage class: %d", sclass);
	  
  switch(type) {
  case type_vector:
    size = sizeof(vector)-4;
    multiplier = 4;
    break;
  case type_structure:
    size = sizeof(structure)-4;
    multiplier = 4;
    break;
  case type_string:
    size = sizeof(string)-1;
    multiplier = 1;
    break;
  default:
    printf("Can not variable allocate type: %d\n", type);
    break;}
	
  ret = rawalloc(size+headersize+multiplier*elements);
  *ret = (int) type;
  return (object) ret;}
	    

static object address_of_proc (object obj) {
  return make_fixnum ((int)obj);}


static char clean(char s) {return s<32?'.':s>126?'.':s;}


static int *printmem(int *mem, int size) {
  int i, j;
  char *ch;

  ch=(char *)mem;

  printf("#x%8.8x: ", (void *)mem);
  for (j=0; j<size; j++)
    printf(" %8.8x", *mem++);
  printf("   ");
  for (j=0; j<size; j++)
    /* display the characters in memory order.  this is different from
       how the 32bit quantities are displayed. */
    {printf(" ");
     for (i=0; i<4; i++)
       printf("%c", clean(*ch++));
      }
  printf("\n");
  return mem;}


static int printchunk(int addr, int totalbytes) {
  int *mem;
  int j;
  mem=(int *)addr;
  for (j=0; j<((totalbytes+15)>>4); j++)
    mem=printmem(mem, 4);
  return totalbytes;}


static object display_memory_proc (object begin, object length) {
  if (! ((fixnump(begin) && (fixnump(length)))))
    error2("Must supply fixnum begin and length, not ~S & ~S", begin, length);
  printchunk((~3)&(fixnum(begin)->value), fixnum(length)->value);
  return nilobj;}

static object objectify (object begin) {
  if (!((fixnump(begin))&&(0==(3&fixnum(begin)->value))))
    error1("Must supply a fixnum (mod 4), not ~S", begin);
  return (object)(fixnum(begin)->value);}

/* allocate various sorts of objects */

static object make_cpointer (void *point) {
  object ret;
  ret = lalloc(type_cpointer, class_object);
  cpointer(ret)->point = point;
  return ret;}
	
	
static object make_fixnum (int num) {
  object ret;
  if ((num>=-fixnumoffset)&&(num<fixnumoffset))
    return fixarray[num+fixnumoffset];
  ret = lalloc(type_fixnum, class_object);
  fixnum(ret)->value = num;
  return ret;}
	

static object make_flonum (double num) {
  object ret;
  ret = lalloc(type_flonum, class_object);
  flonum(ret)->value = num;
  return ret;}
	

static object make_string (char *rawstring) {
  int len;
  char *st, *newst;
  object ret; 
		
  len = 0;
  st = rawstring;
  while (*st++) len++;
  
  ret = lallocvar(type_string, class_object, len);
  string(ret)->length = len;
  newst = string(ret)->chars;
	
  st = rawstring;
  while ((*newst++)=(*st++));

  return ret;}
	

static object make_symbol (char *name, object value, object proc) {
  object ret;
  savereg(value);
  savereg(proc);
  ret = intern(name);
  restorereg(proc);
  restorereg(value);
  symbol(ret)->value = value;
  symbol(ret)->procedure = proc;
  return ret;}
	

static object cons (object car, object cdr) {
  object ret;
  savereg(car);
  savereg(cdr);
  ret = lalloc(type_cons, class_object);
  restorereg(cdr);
  restorereg(car);
  pair(ret)->car = car;
  pair(ret)->cdr = cdr;
  return ret;}


static object make_procedure (int info) {
  object ret;
  ret = lalloc(type_procedure, class_object);
  procedure(ret)->info = info;
  procedure(ret)->proc = nilobj;
  procedure(ret)->arglist = nilobj;
  procedure(ret)->env = nilobj;
  return ret;}


static object make_vector (int length) {
  object ret;
  int i; 
  ret = lallocvar(type_vector, class_object, length);
  vector(ret)->length = length;
  for (i=0;i<length;i++) vector(ret)->arr[i] = nilobj;
  return ret;}


static object make_structure (object symb) {
  object ret, numslots, descriptor, defaults;
  int i, ns, hashindex;
	
  hashindex = get_hashindex(structtable, string(symbol(symb)->name)->chars,
			    hashlooker(struct_desc_name));
  descriptor = vector(structtable)->arr[hashindex];
	
  if (nilobj==descriptor)
    error1("Unknown structure type: ~S", symb);

  numslots = structref(descriptor,struct_numslots);
  ns = fixnum(numslots)->value;
  savereg(descriptor);
  ret = lallocvar(type_structure, class_object, ns);
  restorereg(descriptor);
  structure(ret)->numslots = ns;
  structure(ret)->descriptor = descriptor;
  defaults = structref(descriptor,struct_slotdefaults);
  for(i=0;i<ns;i++)
    structure(ret)->slots[i]
      = (nilobj==defaults)?nilobj:vector(defaults)->arr[i];
  return ret;}


static object struct_desc_name (object descriptor) {
  return symbol(structref(descriptor,struct_type))->name;}

	
static object make_structure_structure () {
  object ret, vec;
  int numslots = 4;
	
  ret = lallocvar(type_structure, class_object, numslots);
  structure(ret)->numslots = numslots;
  structure(ret)->descriptor = ret;
  structref(ret,struct_type) = structuresym = intern("STRUCTURE");
  structref(ret,struct_numslots) = make_fixnum(numslots);
  vec = make_vector(numslots);
  vector(vec)->arr[struct_type] = intern(":type");
  vector(vec)->arr[struct_numslots] = intern(":numslots");
  vector(vec)->arr[struct_slotnames] = intern(":slotnames");
  vector(vec)->arr[struct_slotdefaults] = intern(":slotdefaults");
  structref(ret,struct_slotnames) = vec;
  structref(ret,struct_slotdefaults) = nilobj;
  vector(structtable)->arr[get_hashindex(structtable,
					 string(symbol_name(structuresym))->chars,
					 hashlooker(struct_desc_name))]
    = ret;
  return ret;}


static void register_structure (object descriptor) {
  object desc;
  if ((structurep(descriptor))
      &&(structurep((desc=structure(descriptor)->descriptor)))
      &&(structuresym==structref(desc,struct_type)))
    vector(structtable)->arr[get_hashindex(structtable,
					   string(struct_desc_name(descriptor))->chars,
					   hashlooker(struct_desc_name))]
      = descriptor;
  else
    error1("Can't register a non structure: ~S", descriptor);
  return;}
	

/* external allocation of objects in the heap

   first one must use register_ltype to get allocated a type number
   indicate how it is to be printed out, and tell how big it should be
   allocated.

   then once calls allocate_ltype to get an instance, including the
   type in the header.
*/

static int next_ltype = type_last;

extern int register_ltype (char *descriptor, int size) {
  if (next_ltype>=max_ltypes)
    {printf("\n\nTrying to register %s when we are out of space...\n\n",
	    descriptor);
    exit(-1);}
  vector(ltypes)->arr[next_ltype-type_last]
    = cons(make_string(descriptor),make_fixnum(size));
  return next_ltype++;}


extern int *allocate_ltype (int type) {
  int *ret;
  ret = crawalloc(fixnum(cdr(vector(ltypes)->arr[type-type_last]))->value);
  *ret = (int) type;
  return ret;}


#define ltypep(arg)	((type_last <= gettype(arg))\
                         &&(next_ltype > gettype(arg)))
#define lpointerp(arg)  ((type_last > gettype(arg)))

/* very primitive hash tables

   really just is a flag table as to whether objects are in there
   or not. the *lookup procedure gets a string object out of a table
   entry, and that is what is hashed on.
*/


static int get_hashindex (object vec, char *str, object (*lookup)(object)) {
  int veclength = vector(vec)->length;
  int hash = (compute_hash(str))%veclength;
	
  while (((vector(vec)->arr[hash]) != nilobj)
	 && different_strings(str, (*lookup)(vector(vec)->arr[hash])))
    hash = (hash+1)%veclength;
  return hash;}


static int compute_hash (char *str) {
  char *st = str;
  int hash = 0;
  int ch;
	
  while (ch=*st++) hash = hash ^ (ch+(hash<<2));
  return hash&0xfffffff;}


static boolean different_strings (char *str1, object str2) {
  int i, len;
  char *st1, *st2;

  len = 1+string(str2)->length;
  st1 = str1;
  st2 = string(str2)->chars;
  for (i=0; i<len; i++)
    if ((*st1++) != (*st2++))
      return true;
  return false;}
	 
	
/* handle interning a symbol */

static object symbol_name (object symb) {return symbol(symb)->name;}


static object intern (char *str) {
  object newsym, name;
  int index = get_hashindex(obtable, str, hashlooker(symbol_name));
  
  if ((vector(obtable)->arr[index])==nilobj) {
    name = make_string(str);
    savereg(name);
    newsym = lalloc(type_symbol, class_object);
    restorereg(name);
    symbol(newsym)->name = name;
    /* handle the fake keywords... */
    symbol(newsym)->value = (':'==*(string(name)->chars))?newsym:unbound_value;
    symbol(newsym)->procedure = undef_proc;
    vector(obtable)->arr[index] = newsym;}
  
  return (object) vector(obtable)->arr[index];}


/* record of all specials --- things can only be globally special,
   not locally special.
*/

static void declare_special (object sym) {
  if (symbolp(sym))
    vector(specials)->arr[get_hashindex(specials,
					string(symbol_name(sym))->chars,
					hashlooker(symbol_name))]
      = sym;
  else
    error1("~S is not a symbol; can't be declared special", sym);
  return;}


static object is_special_p (object sym) {
  return vector(specials)->arr[get_hashindex(specials,
					     string(symbol_name(sym))->chars,
					     hashlooker(symbol_name))];}


static void declare_constant (object sym) {
  if (symbolp(sym))
    vector(constants)->arr[get_hashindex(constants,
					 string(symbol_name(sym))->chars,
					 hashlooker(symbol_name))]
      = sym;
  else
    error1("~S is not a symbol; can't be declared constant", sym);
  return;}


static object is_constant_p (object sym) {
  if (':'==string(symbol(sym)->name)->chars[0])
    return tsymbol;
  else
    return vector(constants)->arr[get_hashindex(constants,
						string(symbol_name(sym))->chars,
						hashlooker(symbol_name))];}


/* handle installation of c procedures */

extern void register_cprocedure (char *name,
				 void *proc,
				 ctype result,
				 ctype arg0, ctype arg1, ctype arg2) {	
  object sym, tproc;
  sym = intern(name);
  tproc = make_procedure(make_procinfo(proc_cproc,((arg0==c_void)?make_arginfo(0,0):(arg1==c_void)?make_arginfo(1,1):(arg2==c_void)?make_arginfo(2,2):make_arginfo(3,3))));
  symbol(sym)->procedure = tproc;
  procedure(tproc)->name = sym;
  procedure(tproc)->proc = proc;
  procedure(tproc)->arglist = (int *)((result<<24)|(arg0<<16)|(arg1<<8)|arg2);
  return;}


/* give access to FELL variables, and return their values */

extern int check_fixnump (char *name) {
  return fixnump(symbol(intern(name))->value);}

extern int lookup_fixnum (char *name) {
  return fixnum(symbol(intern(name))->value)->value;}

extern int check_flonump (char *name) {
  return flonump(symbol(intern(name))->value);}

extern double lookup_flonum (char *name) {
  return flonum(symbol(intern(name))->value)->value;}

/* give access to FELL cons cells to do list operations externally */

extern boolean lnullp (void *obj) {return nullp((object)obj);}
extern void *lcar (void *obj) {return carproc((object)obj);}
extern void *lcdr (void *obj) {return cdrproc((object)obj);}


/*****************************************************************************
******************************************************************************
******************************************************************************

*** the garbage collector

******************************************************************************
******************************************************************************
*****************************************************************************/

/* this is a straigtforward stop and copy garbage collector.
   the trick is getting all the roots out of the middle of the
   evaluator....

   staticspace is certainly all a root.
*/


static boolean gcprintflag = false;

static object gcprint (object flag) {
  gcprintflag = (flag!=nilobj);
  return flag;}


static object gc () {
  memspace temp;
  int i, *mem;

  temp = tospace;
  tospace = fromspace;
  fromspace = temp;

  mem = (int *)tospace->lowest;
  for(i=(tospace->size)<<2;i>0;i--) *mem=0;
  tospace->freepoint = tospace->lowest;
  tospace->freemem = tospace->size;

  gc_sweep(staticspace, fromspace);
  for(i=0;i<=sp;i++) stack[i] = gc_forward(stack[i],fromspace);
  for(i=0;i<=regsp;i++) regstack[i] = gc_forward(regstack[i],fromspace);
  gc_sweep(tospace, fromspace);

  if (gcprintflag)
    printf("\n;;; GC produced %d free bytes.\n", tospace->freemem);
  gc_count++;
  return nilobj;}


#define roundsize(structname)  ((~3)&(3+headersize+sizeof(structname)))
#define forwardref(place)      place=gc_forward(place,from)

static void gc_sweep (memspace sweepspace, memspace from) {
  object obj;
  int len, i;
  lisptype type;
  char *sweep;
  
  sweep = sweepspace->lowest;
  /* in the case of tospace the freepoint may change... */
  while (sweep<sweepspace->freepoint){
    obj = (object)sweep;
    type = gettype(obj);
    switch (type) {
    case type_null: case type_cons:
      forwardref(pair(obj)->car);
      forwardref(pair(obj)->cdr);
      sweep += roundsize(pair);
      break;
    case type_fixnum:
      sweep += roundsize(fixnum);
      break;
    case type_flonum:
      sweep += roundsize(flonum);
      break;
    case type_string:
      sweep += (~3)&(3+headersize+sizeof(string)-1+string(obj)->length);
      break;
    case type_symbol:
      forwardref(symbol(obj)->name);
      forwardref(symbol(obj)->value);
      forwardref(symbol(obj)->procedure);
      sweep += roundsize(symbol);
      break;
    case type_vector:
      len = vector(obj)->length;
      for(i=0;i<len;i++) forwardref(vector(obj)->arr[i]);
      sweep += (~3)&(3+headersize+sizeof(vector)-4+4*len);
      break;
    case type_structure:
      len = structure(obj)->numslots;
      forwardref(structure(obj)->descriptor);
      for(i=0;i<len;i++) forwardref(structure(obj)->slots[i]);
      sweep += (~3)&(3+headersize+sizeof(structure)-4+4*len);
      break;
    case type_procedure:
      forwardref(procedure(obj)->name);
      forwardref(procedure(obj)->arglist);
      forwardref(procedure(obj)->env);
      forwardref(procedure(obj)->proc);
      sweep += roundsize(procedure);
      break;
    case type_marker:
      sweep += roundsize(marker);
      break;
    case type_cpointer:
      sweep += roundsize(cpointer);
      break;
    case type_filler:
      sweep += headersize;
      break;
    default:
      printf("\nDon't understand this in GC: %8.8x\n", type);
      exit(-1);}
  }

  return;}


static object gc_forward (object obj, memspace from) {
  int fromlo, fromhi, cand, len, i;
  char *newst, *oldst;
  int *newvc, *oldvc;
  object ret;
  fromlo = (int)from->lowest;
  fromhi = from->size+fromlo;
  if ((((int)obj)>=fromlo) && (((int)obj)<fromhi)) {
    cand = gettype(obj);
    if (cand==(cand&0xff)) {
      switch(cand) {
      case type_fixnum: return (object)(*obj)=make_fixnum(fixnum(obj)->value);
      case type_flonum: return (object)(*obj)=make_flonum(flonum(obj)->value);
      case type_cons: return (object)(*obj)=cons(car(obj), cdr(obj));
      case type_string:
	len = string(obj)->length;
	ret = lallocvar(type_string, class_object, len);
	string(ret)->length = len;
	newst = string(ret)->chars;
	oldst = string(obj)->chars;
	while (*newst++=*oldst++);
	return (object)(*obj)=ret;
      case type_symbol:
	ret = lalloc(type_symbol, class_object);
	symbol(ret)->name = symbol(obj)->name;
	symbol(ret)->value = symbol(obj)->value;
	symbol(ret)->procedure = symbol(obj)->procedure;
	return (object)(*obj)=ret;
      case type_vector:
	len = vector(obj)->length;
	ret = lallocvar(type_vector, class_object, len);
	vector(ret)->length = len;
	for(i=0;i<len;i++) vector(ret)->arr[i] = vector(obj)->arr[i];
	return (object)(*obj)=ret;
      case type_structure:
	len = structure(obj)->numslots;
	ret = lallocvar(type_structure, class_object, len);
	structure(ret)->numslots = len;
	structure(ret)->descriptor = structure(obj)->descriptor;
	for(i=0;i<len;i++) structure(ret)->slots[i] = structure(obj)->slots[i];
	return (object)(*obj)=ret;
      case type_procedure:
	ret = lalloc(type_procedure, class_object);
	procedure(ret)->info = procedure(obj)->info;
	procedure(ret)->proc = procedure(obj)->proc;
	procedure(ret)->arglist = procedure(obj)->arglist;
	procedure(ret)->name = procedure(obj)->name;
	procedure(ret)->env = procedure(obj)->env;
	return (object)(*obj)=ret;
      case type_marker: 
	ret = lalloc(type_marker, class_object);
	marker(ret)->mark = marker(obj)->mark;
	return (object)(*obj)=ret;
      case type_cpointer:
	return (object)(*obj)=make_cpointer(cpointer(obj)->point);
      default:
	printf("\nCan not forward an object of type: %d\n", cand);
	exit(-1);}}
    else
      return (object)cand;}
  else
    return obj;}


/*****************************************************************************
******************************************************************************
******************************************************************************

*** the printer

******************************************************************************
******************************************************************************
*****************************************************************************/

char printbuffer[1000];
#define printbufsize 999

#define forcettystream(stream)	if (stream==nilobj) stream = tsymbol


static object print (object obj, object stream) {
  pmethod method;
  forcettystream(stream);
  method = get_print_method(stream);
  outchar('\n', method);
  prin1_aux(obj, method);
  outchar(' ', method);
  force_output(method);
  return obj;}


static object prin1 (object obj, object stream) {
  pmethod method;
  forcettystream(stream);
  method = get_print_method(stream);
  prin1_aux(obj, method);
  force_output(method);
  return obj;}


static object princ (object obj, object stream) {
  pmethod method;
  forcettystream(stream);
  method = get_print_method(stream);
  princ_aux(obj, method);
  force_output(method);
  return obj;}


static object terpri (object stream) {
  pmethod method;
  forcettystream(stream);
  method = get_print_method(stream);
  outchar('\n', method);
  force_output(method);
  return nilobj;}


static object printproc (object obj, object restargs) {
  return print(obj,nilobj==restargs?tsymbol:first(restargs));}

static object prin1proc (object obj, object restargs) {
  return prin1(obj,nilobj==restargs?tsymbol:first(restargs));}

static object princproc (object obj, object restargs) {
  return princ(obj,nilobj==restargs?tsymbol:first(restargs));}

static object terpriproc (object restargs) {
  return terpri(nilobj==restargs?tsymbol:first(restargs));}

static object format (object stream, object strng, object restargs) {
  pmethod method;
  int len, i, j, numeric, num, tabplace, floatsize;
  double fnum;
  char *str, ch, leader;
  boolean numericflag, floatsizeflag;
	
  if (!stringp(strng)) error1("FORMAT non-string: ~S", strng);
  str = string(strng)->chars;
  len = string(strng)->length;
  i = 0;
  method = get_print_method(stream);
  goto endtest;
 error:
  error1("Unparsable FORMAT string: ~S", strng);
  goto done;
 shorterr:
  error1("Too few arguments to FORMAT: ~S", strng);
  goto done;
 loop:
  ch = str[i];
  if (ch=='~')
    {numericflag = false;
    numeric = 0;
    floatsizeflag = false;
    floatsize = 0;
    leader = ' ';
    innerloop:
    if ((++i)>=len) goto error;
    ch = str[i];
    switch(ch){
    case '%': ch = '\n'; goto outchar;
    case '&': if (numericflag && (0==numeric)) goto more;
      if (0!=method->column) outchar('\n', method);
      if (numericflag)
	{numeric = (numeric>1)?numeric-1:0;
	ch = '\n';
	goto outchar;}
      goto more;
    case '~': goto outchar;
    case 'T': case 't': outtab(numericflag, numeric, method); goto more;
    case 'D': case 'd': if (nilobj==restargs) goto shorterr;
      if (fixnump(first(restargs)))
	num = fixnum(first(restargs))->value;
      else
	error1("Expecting a fixnum, not ~S", first(restargs));
      prin1_fixnum_main_c(num, method, numericflag?numeric:-1, leader, 10);
      restargs = rest(restargs);
      goto more;
    case 'B': case 'b': if (nilobj==restargs) goto shorterr;
      if (fixnump(first(restargs)))
	num = fixnum(first(restargs))->value;
      else
	error1("Expecting a fixnum, not ~S", first(restargs));
      prin1_fixnum_main_c(num, method, numericflag?numeric:-1, leader, 2);
      restargs = rest(restargs);
      goto more;
    case 'O': case 'o': if (nilobj==restargs) goto shorterr;
      if (fixnump(first(restargs)))
	num = fixnum(first(restargs))->value;
      else
	error1("Expecting a fixnum, not ~S", first(restargs));
      prin1_fixnum_main_c(num, method, numericflag?numeric:-1, leader, 8);
      restargs = rest(restargs);
      goto more;
    case 'X': case 'x': if (nilobj==restargs) goto shorterr;
      if (fixnump(first(restargs)))
	num = fixnum(first(restargs))->value;
      else
	error1("Expecting a fixnum, not ~S", first(restargs));
      prin1_fixnum_main_c(num, method, numericflag?numeric:-1, leader, 16);
      restargs = rest(restargs);
      goto more;
    case 'F': case 'f': if (nilobj==restargs) goto shorterr;
      if (flonump(first(restargs)))
	fnum = flonum(first(restargs))->value;
      else
	error1("Expecting a flonum, not ~S", first(restargs));
      prin1_float_size(fnum,
		       method,
		       numericflag?numeric:-1,
		       floatsizeflag?floatsize:-1);
      restargs = rest(restargs);
      goto more;
    case 'S': case 's':
    case 'A': case 'a': if (nilobj==restargs) goto shorterr;
      tabplace = numericflag?numeric+method->column:-1;
      if ((ch=='S')||(ch=='s'))
	prin1_aux(first(restargs), method);
      else
	princ_aux(first(restargs), method);
      if ((tabplace>-1)&&(method->column < tabplace))
	outtab(true, tabplace, method);
      restargs = rest(restargs);
      goto more;
    case ',': i = i+2;
      if ((len>i+1)
	  &&('\''==str[i-1])
	  &&(('D'==str[i+1])||('d'==str[i+1])
	     ||('X'==str[i+1])||('x'==str[i+1])
	     ||('O'==str[i+1])||('o'==str[i+1])
	     ||('B'==str[i+1])||('b'==str[i+1])))
	leader = str[i];
      else if ((len>i-1)
	       &&(('F'==str[i-1])||('f'==str[i-1])))
	{floatsizeflag = false;
	i = i-2;}
      else if ((len>i)
	       &&(('F'==str[i])||('f'==str[i]))
	       &&('1'<=str[i-1])&&(str[i-1]<='9'))
	{i = i-1;
	floatsizeflag = true;
	floatsize = str[i]-'0';}
      else goto error;
      break;
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      numeric = (10*(numericflag?numeric:0))+ch-'0';
      numericflag = true;
      break;
    default: goto error;
    }
    goto innerloop;
    outchar:
    for (j=0;j<(numericflag?numeric:1);j++)
      outchar(ch, method);
    }
  else
    outchar(ch, method);
 more:
  i++;
 endtest:
  if (i<len) goto loop;
 done:
  if (nilobj==stream)
    {method->string[method->count]=0;
    return make_string(method->string);}
  else
    {force_output(method);
    return nilobj;}}
               
             
static void outtab(boolean flag, int number, pmethod method) {
  int column, i, range;
	
  column = method->column;
  range = (!flag)?1:(column<number)?number-column:1;
  for (i=0; i<range; i++)
    outchar(' ', method);
  return;}


/* the worker internals for printing */
            
static void princ_aux (object obj, pmethod method) {
  switch (gettype(obj)){
  case type_string: princ_string_c(string(obj)->chars, method); break;
  case type_symbol:
    princ_string_c(string(symbol(obj)->name)->chars, method);
    break;
  default: prin1_aux(obj, method); break;
  }
  return;}


static void prin1_aux (object obj, pmethod method) {
  int saved_plevel, plength, vlength, i, type;
  object levelobj, pstructobj, snamevec;
  
  switch (type=gettype(obj)){
  case type_null: princ_string_c("()", method); break;
  case type_flonum: prin1_float_c(flonum(obj)->value, method); break;
  case type_fixnum:
    prin1_fixnum_main_c(fixnum(obj)->value,
			method,
			-1,
			' ',
			fixnum(symbol(printbase)->value)->value);
    break;
  case type_string:
    outchar('"', method);
    princ_string_c(string(obj)->chars, method);
    outchar('"', method);
    break;
    /* may include packages here later... */
  case type_symbol:
    princ_string_c(string(symbol(obj)->name)->chars, method);
    break;
  case type_cons:
    levelobj = symbol(printlevel)->value;
    if ((nilobj!=levelobj) && (fixnum(levelobj)->value < 1))
      outchar('#', method);
    else if ((quote == car(obj))
	     && (type_cons == gettype(cdr(obj)))
	     && nilobj == cdr(cdr(obj)))
      {outchar('\'', method);
      prin1_aux(car(cdr(obj)), method);}
    
    else if ((function == car(obj))
	     && (type_cons == gettype(cdr(obj)))
	     && nilobj == cdr(cdr(obj)))
      {outchar('#', method);
      outchar('\'', method);
      prin1_aux(car(cdr(obj)), method);}
	  		   
    else
      {outchar('(', method);
      levelobj = symbol(printlevel)->value;
      saved_plevel = (nilobj==levelobj)?-1:fixnum(levelobj)->value;
      if (nilobj!=levelobj) fixnum(levelobj)->value = saved_plevel -1;
      plength = (nilobj==symbol(printlength)->value)
	?1000000
	:fixnum(symbol(printlength)->value)->value;
      i = 1;
      prin1_aux(car(obj), method);
      while (type_cons==gettype((obj=cdr(obj))))
	{outchar(' ', method);
	if (i>=plength)
	  {princ_string_c("...", method);
	  goto done;}
	prin1_aux(car(obj), method);
	i++;}
      if (type_null != gettype(obj))
	{princ_string_c(" . ", method);
	if (i>=plength)
	  princ_string_c("...", method);
	else
	  prin1_aux(obj, method);}
      done:
      if (nilobj!=levelobj)
	fixnum(levelobj)->value = saved_plevel;
      outchar(')', method);}
    break;
  case type_vector:
    levelobj = symbol(printlevel)->value;
    if ((levelobj!=nilobj)&&(fixnum(levelobj)->value < 1))
      outchar('#', method);
    else if (symbol(printarray)->value!=nilobj)
      {saved_plevel = (nilobj==levelobj)?-1:fixnum(levelobj)->value;
      if (nilobj!=levelobj) fixnum(levelobj)->value = saved_plevel -1;
      outchar('#', method);
      outchar('(', method);
      plength = (nilobj==symbol(printlength)->value)
	?1000000
	:fixnum(symbol(printlength)->value)->value;
      vlength = vector(obj)->length;
      for (i=0;i<vlength;i++)
	{if (i!=0) outchar(' ', method);
	if (i>plength)
	  {princ_string_c("...", method);
	  break;}
	prin1_aux(vector(obj)->arr[i], method);}
      outchar(')', method);
      if (nilobj!=levelobj)
	fixnum(levelobj)->value = saved_plevel;}
    else
      {princ_string_c("#<vector (", method);
      prin1_fixnum_main_c(vector(obj)->length, method, -1, ' ', 10);
      princ_string_c("): #x", method);
      prin1_fixnum_main_c((int)obj, method, 8, '0', 16);
      outchar('>', method);}
    break;
  case type_structure:
    pstructobj = symbol(printstructure)->value;
    levelobj = symbol(printlevel)->value;
    if ((levelobj!=nilobj)&&(fixnum(levelobj)->value < 1))
      outchar('#', method);
    else if (pstructobj!=nilobj)
      {saved_plevel = (nilobj==levelobj)?-1:fixnum(levelobj)->value;
      if (nilobj!=levelobj) fixnum(levelobj)->value = saved_plevel -1;
      princ_string_c("#S(", method);
      prin1_aux(structref(structure(obj)->descriptor, struct_type), method);
      outchar(' ', method);
      plength = (nilobj==symbol(printlength)->value)?1000000:fixnum(symbol(printlength)->value)->value;
      vlength = structure(obj)->numslots;
      snamevec = structref(structure(obj)->descriptor, struct_slotnames);
      for (i=0;i<vlength;i++)
	{if (i!=0) outchar(' ', method);
	if (i>plength)
	  {princ_string_c("...", method);
	  break;}
	prin1_aux(vector(snamevec)->arr[i], method);
	outchar(' ', method);
	prin1_aux(structure(obj)->slots[i], method);}
      outchar(')', method);
      if (nilobj!=levelobj)
	fixnum(levelobj)->value = saved_plevel;}
    else
      {princ_string_c("#<", method);
      prin1_aux(structref(structure(obj)->descriptor, struct_type) ,method);
      princ_string_c(" #x", method);
      prin1_fixnum_main_c((int)obj, method, 8, '0', 16);
      outchar('>', method);}
    break;
  case type_marker:
    outchar('#', method);
    outchar('<', method);
    princ_string_c((marker(obj)->mark==mark_unbound_value)?"unbound":
		   (marker(obj)->mark==mark_undef_proc)?"undefined":"?????",
		   method);
    outchar('>', method);
    break;
  case type_procedure:
    princ_string_c("#<procedure ", method);
    prin1_aux(procedure(obj)->name, method);
    princ_string_c(" #x", method);
    prin1_fixnum_main_c((int)obj, method, 8, '0', 16);
    outchar('>', method);
    break;
  case type_cpointer:
    princ_string_c("#<C pointer #x", method);
    prin1_fixnum_main_c((int)cpointer(obj)->point, method, 8, '0', 16);
    outchar('>', method);
    break;
	  		  	
  default:
    if ((type_last <= type)&&(next_ltype > type))
      {princ_string_c("#<", method);
      princ_string_c(string(car(vector(ltypes)->arr[type-type_last]))->chars, method);
      outchar(' ', method);
      princ_string_c(getltypestring(obj), method);
      princ_string_c(" #x", method);}
    else
      princ_string_c("#<unprintable #x", method);
    prin1_fixnum_main_c((int)obj, method, 8, '0', 16);
    outchar('>', method);
    break;}
  return;}
	

static void princ_string_c (char *str, pmethod method) {
  char *st, ch;
    
  st = str;
  while (ch=*st++) outchar(ch, method);
  return;}


static char floatstring[40];
  
static void prin1_float_c (double f, pmethod method) {
  int i, c;
  boolean point;

  point = false;
  c = sprintf(floatstring, "%g", f);
  for (i=0;i<c;i++)
    if (floatstring[i]=='.') {
      point = true;
      break;}
  if (!point)
    {floatstring[c++]='.';
    floatstring[c++]='0';}
  for (i=0;i<c;i++) outchar(floatstring[i], method);
  return;}
    

static void prin1_float_size (double number, pmethod method, int size, int places) {
  int c, i;
  if (places==-1) {
    c = sprintf(floatstring, "%*f", size, number);}
  else {
    c = sprintf(floatstring, "%*.*f", size, places, number);}
  for (i=0;i<c;i++) outchar(floatstring[i], method);
  return;}


static void prin1_fixnum_main_c (int num, pmethod method, int trail, char trailchar, int base) {
  int i;
  if (0==num)
    {if (trail>1)
      for (i=1; i<trail; i++) outchar(trailchar, method);
    outchar('0', method);}
  else if (num>0)
    prin1_fixnum_c(num, method, trail, trailchar, base);
  else if (trail==-1)
    {outchar('-' , method);
    prin1_fixnum_c((0-num), method, trail, ' ', base);}
  else
    prin1_fixnum_c((0-num), method, trail, '-', base);
  return;}
	

static void prin1_fixnum_c (int num, pmethod method, int trail, char trailchar, int base) {
  int digit, others;
  if (0==trail)
    return;
  else if ((1==trail) && (trailchar == '-'))
    outchar('-', method);
  else
    {digit = num%base;
    others = num/base;
    if ((others!=0) || (trail > 0))
      prin1_fixnum_c(others,
		     method,
		     (trail>0)?(trail-1):-1,
		     ((0==num) && (trailchar == '-'))?' ':trailchar,
		     base);
    outchar((0==num)?trailchar:(digit<10)?(digit+'0'):('A'+digit-10), method);}
  return;}


/* a pmethod has the following slots:
   :count = number of characters in buffer
   :string = a character buffer
   :size = length of the output buffer
   :column = next column that a character will be printed on
   :stream = ()	return a string as the result of the printing (format only)
   	     t	use the tty
   	     <stream> an open output file
*/


static pmethod new_pmethod() {
  pmethod ret;
  ret = (pmethod)malloc(20);
  ret->count = 0;
  ret->column = 0;
  ret->size = 0;
  ret->stream = tsymbol;
  ret->string = NULL;
  return ret;}


static void initialize_printer() {
  tprintmethod = new_pmethod();		
  return;}
	

static pmethod get_print_method (object stream) {
  pmethod ret;
  if (tsymbol==stream)
    return tprintmethod;
  else if (nilobj==stream)
    {ret = new_pmethod();
    ret->string = printbuffer;
    ret->size = printbufsize - 1;
    ret->stream = nilobj;
    return ret;}
  else if (pairp(stream) && (first(stream)==filekey)
	   && (second(stream) == outputkey)
	   && (((lstream)(rest(rest(stream))))->openstatus == true))
    return ((lstream)rest(rest(stream)))->prmeth;
  else
    error0("Unimplemented print method");}


static void outchar (char ch, pmethod method) {
  object stream;
  stream = (object)method->stream;
  if (ch=='\n')
    method->column=0;
  else
    method->column++;
  if (tsymbol==stream)
    putchar(ch);
  else if (nilobj==stream)
    {method->string[method->count]=ch;
    method->count++;
    if (method->count >= method->size)
      error0("Overflow in format to string");}
  else
    putc(ch, ((lstream)stream)->openfile);
  return;}


static void force_output (pmethod method) {
  return;}


/*****************************************************************************
******************************************************************************
******************************************************************************

*** the reader

******************************************************************************
******************************************************************************
*****************************************************************************/


static boolean reader_internal = false;
static int backquote_level = 0;
static lstream instream = NULL;

static lstream stringlstream;
static lstream ttylstream;

static rcatchpoint *rparen_catch = NULL;
static rcatchpoint *dot_catch = NULL;
static rcatchpoint *dot_error_catch = NULL;
static rcatchpoint *eof_catch = NULL;

#define rcatchpoints (200)
static rcatchpoint rcatches[rcatchpoints];

static int freercatch = -1;

static rcatchpoint *get_freercatchpoint () {
  rcatchpoint *ret;
  if (rcatchpoints<=freercatch++)
    {printf("\n\nRan out of reader catch points!!!\n\n");
    exit(-1);}
  ret = &rcatches[freercatch];
  ret->readinternal = reader_internal;
  ret->bqlevel = backquote_level;
  ret->rparen = rparen_catch;
  ret->dot = dot_catch;
  ret->doterror = dot_error_catch;
  ret->eof = eof_catch;
  ret->free = freercatch-1;
  return ret;}
	
static void restore_rcatchpoint (rcatchpoint *catchp) {
  reader_internal = catchp->readinternal;
  backquote_level = catchp->bqlevel;
  rparen_catch = catchp->rparen;
  dot_catch = catchp->dot;
  dot_error_catch = catchp->doterror;
  eof_catch = catchp->eof;
  freercatch = catchp->free;
  return;}
	 

/* the table describing all the characters in the character set */

static char reader_table[128];

#define maxtoken 128
static char reader_buff[maxtoken];

#define lconstituent 1
#define lwhitespace 2
#define ltermmacro 4
#define lnontermmacro 8
#define lsingleescape 16
#define lmultipleescape 32
#define lalphabetic 64

static object openkeywords;
static object readfromstringkeywords;

static void init_reader () {
  int i;
  char *str, ch;
	
  for (i=0;i<128;i++) reader_table[i]=0;
  for (i='0';i<='9';i++) reader_table[i] = lconstituent;
  for (i='a';i<='z';i++) reader_table[i] = lconstituent|lalphabetic;
  for (i='A';i<='Z';i++) reader_table[i] = lconstituent|lalphabetic;
  reader_table['\\'] = lsingleescape;
  reader_table['|'] = lmultipleescape;
  reader_table['#'] = lnontermmacro;
  str = "!$%&*+-./:<=>?@[]^_{}~";
  while (ch=*str++) reader_table[ch] = lconstituent;
  str = "'(),;`\"";
  while (ch=*str++) reader_table[ch] = ltermmacro;
  str = " \t\a\f\n\r\v";
  while (ch=*str++) reader_table[ch] = lwhitespace;
    
  stringlstream = malloc(sizeof(streamstruct));
  ttylstream = malloc(sizeof(streamstruct));
  openkeywords = list2(directkey,ifexistskey);
  readfromstringkeywords = list2(startkey, endkey);
  
  return;}
	

/* some internal data structures and macros for the reader. */

typedef enum {notoken, token} tokenindicator;

#define whitespacep(attr)	(0 != (lwhitespace & attr))
#define macrop(attr)		(0 != ((ltermmacro | lnontermmacro)&attr))
#define termmacrop(attr)	(0 != (ltermmacro&attr))
#define escapep(attr)		(0 != ((lsingleescape | lmultipleescape)&attr))
#define singleescapep(attr)	(0 != (lsingleescape&attr))
#define tokencharp(attr)	(0 != ((lnontermmacro | lconstituent)&attr))


/* the reader itself */

static object fellread (object restargs) {
  object stream, eof_error_p, eof_value;
	
  stream = tsymbol;
  eof_error_p = tsymbol;
  eof_value = nilobj;
	
  if (pairp(restargs))
    {stream = first(restargs);
    restargs = rest(restargs);
    if (pairp(restargs))
      {eof_error_p = first(restargs);
      restargs = rest(restargs);
      if (pairp(restargs))
	eof_value = first(restargs);}}
  return readproc(stream, eof_error_p, eof_value);}
	
	
/* this is not GC sensitive.  so it forces there to be 100,000 bytes
   free before getting started and hopes for the best.
*/

static object readproc (object stream, object eof_error_p, object eof_value) {
  int saved_bqlevel;
  int savedspace;
  boolean saved_rinternal;
  lstream saved_instream;
  
  rcatchpoint *savedpoint;
  object ret;
  
  if ((100000>tospace->freemem)&&(tospace!=staticspace)) {
    savereg(stream);
    savereg(eof_error_p);
    savereg(eof_value);
    gc();
    restorereg(eof_value);
    restorereg(eof_error_p);
    restorereg(stream);}

  savedspace = (int)tospace->lowest;

  saved_bqlevel = backquote_level;
  saved_rinternal = reader_internal;
  saved_instream = instream;
  
  reader_internal = false;
  backquote_level = 0;
  instream = get_read_method(stream);
  
  savedpoint = eof_catch = get_freercatchpoint();
  dot_catch = dot_error_catch = get_freercatchpoint();
  rparen_catch = get_freercatchpoint();
  
  if (0 != setjmp(eof_catch->env))
    {ret = end_of_file_error(instream, eof_error_p, eof_value);
    goto cleanup;}
  
  if (0 != setjmp(dot_catch->env))
    {ret = nilobj;
    read_general_error("Dot context");
    goto cleanup;}
  
  if (0 != setjmp(rparen_catch->env))
    {ret = nilobj;
    read_general_error("Unmatched right paren");
    goto cleanup;}
  
  ret = read_aux();
  
 cleanup:
  restore_rcatchpoint(savedpoint);
  reader_internal = saved_rinternal;
  backquote_level = saved_bqlevel;
  instream = saved_instream;

  if (savedspace!=(int)tospace->lowest) {
    printf("\nHad a GC during read---you lose....\n");
    exit(-1);}

  return ret;}


static lstream get_read_method (object stream) {
  lstream ret;
	
  if (tsymbol==stream)
    {ret = ttylstream;
    ret->lastchar = -1;
    ret->openstatus = true;
    ret->openfile = (void *)-1;}
  else if (stringkeyword==stream)
    ret = stringlstream;
  else if (pairp(stream)
	   &&(first(stream)==filekey)
	   &&pairp(rest(stream)))
    {if (second(stream)==inputkey)
      {ret = (lstream)rest(rest(stream));
      if (!ret->openstatus)
	error1("Can not read from a closed file: ~S", stream);
      return ret;}
    else error1("Can not read from an output file: ~S", stream);}
  else error1("Not yet implemented read method for ~S", stream);
  return ret;}


static object read_from_string (object str,
				object eof_error_p,
				object eof_value,
				int start,
				int end) {
  lstream stream = stringlstream;
	
  stream->filename = string(str)->chars;
  stream->openfile = NULL;
  stream->current = start-1;
  stream->limit = end;
  stream->lastchar = -1;
  stream->openstatus = true;
  return readproc(stringkeyword, eof_error_p, eof_value);}


static object read_from_string_proc (object str, object restargs) {
  object eof_error_p, eof_value, finder, cand;
  int start, end;
	
  if (!stringp(str))
    error1("Need a string to read from: ~S", str);
	  
  eof_error_p = tsymbol;
  eof_value = nilobj;
	
  if (nilobj!=restargs)
    {eof_error_p = first(restargs);
    restargs = rest(restargs);}
  if (nilobj!=restargs)
    {eof_value = first(restargs);
    restargs = rest(restargs);}
  checkkeywordargs(restargs, 4, readfromstringkeywords);
	
  start = 0;
  end = string(str)->length;
	
  finder = memq(startkey, restargs);
  if (finder!=nilobj)
    {cand = second(finder);
    if (!fixnump(cand))
      error1("Need a fixnum as a read start: ~S", cand);
    start = fixnum(cand)->value;}
  finder = memq(endkey, restargs);
  if (finder!=nilobj)
    {cand = second(finder);
    if (!fixnump(cand))
      error1("Need a fixnum as a read end: ~S", cand);
    end = fixnum(cand)->value;}
	
  if ((start < 0) || (start > string(str)->length))
    error1("Invalid :start in ~S", restargs);
  
  if ((end < 0) || (end < start) || (end > (string(str)->length)))
    error1("Invalid :end in ~S", restargs);
  
  return read_from_string (str, eof_error_p, eof_value, start, end);}


static object read_aux () {
  char ch;
  int attr, count, i;
  rcatchpoint *thisrparen, *thisdot;
  object list, last, val, newthing, ret;
	
 startread:
  ch = nextchar(notoken);
	
  attr = reader_table[ch];
	
  if (attr==0)
    {read_error("Illegal character code", ch);
    goto startread;}
  else if (whitespacep(attr))
    goto startread;
  else if (macrop(attr))
    /* handling: ();"'`#  */
    {if (ch=='(')
      {reader_internal=true;
      list = nilobj;
      last = nilobj;
      
      goto accumstart;
      
      /* can only exit from below, via longjmp to above */   
      accumlist:
      restore_rcatchpoint(thisrparen);
      accumstart:
      rparen_catch = thisrparen = get_freercatchpoint();
      if (0 != setjmp(thisrparen->env))
	{restore_rcatchpoint(thisrparen);
	return list;}
      
      dot_catch = thisdot = get_freercatchpoint();
      if (0 == setjmp(thisdot->env))
	val = read_aux();
      else
	{restore_rcatchpoint(thisdot);
	if (list == nilobj)
	  longjmp(dot_error_catch->env, 1);
	else
	  {cdr(last)=read_aux();
	  read_aux();
	  longjmp(dot_error_catch->env, 1);}
	}
      restore_rcatchpoint(thisdot);
      if (list == nilobj)
	list = last = cons(val,nilobj);
      else
	{newthing = cons(val,nilobj);
	cdr(last) = newthing;
	last = newthing;}
      goto accumlist;
      }
    else if (ch==')')
      longjmp(rparen_catch->env, 1);
    else if (ch=='#')
      return read_pound_sign();
    else if (ch=='\'')
      return cons(quote,cons(read_aux(),nilobj));
    else if (ch=='`')
      return read_backquote();
    else if (ch=='"')
      return read_string();
    else if (ch==',')
      return read_comma();
    else if (ch==';')
      {while ('\n' != nextchar(notoken)) ;
      goto startread;}
    else
      {read_error("Inexplicable error in READ", ch);
      goto startread;}
    }
  else if (escapep(attr))
    {read_lerror(ch,attr);
    goto startread;}
  else
    {count=0;
    accum:
    if (count==maxtoken-1)
      {reader_buff[maxtoken-1]=0;
      read_error_long();
      goto startread;}
    reader_buff[count] = ch;
    count++;
    ch = nextchar(token);
    if (0 != (0x80 & ch)) goto gottoken;
    attr = reader_table[ch];
    if (tokencharp(attr))
      goto accum;
    else if (whitespacep(attr))
      goto gottoken;
    else if (termmacrop(attr))
      {backchar(ch);
      goto gottoken;}
    else if (escapep(attr))
      read_lerror(ch, attr);
    else
      read_error("Illegal character code", ch);
    gottoken:
    if ((count==1)&&(reader_buff[0]=='.'))
      longjmp(dot_catch->env, 1);
    for (i=0;i<count;i++)
      if (reader_buff[i] != '.') goto nodots;
    longjmp(dot_error_catch->env, 1);
    nodots:
    reader_buff[count]=0;
    ret = read_is_number_p(reader_buff);
    return (ret==nilobj)?intern(reader_buff):ret;
    }}
	 

static object read_string() {
  int count = 0;
  int ch;
  
 more:
  ch = nextchar(notoken);
  if (ch == '\"') goto done;
  if (ch == '\\') ch = nextchar(notoken);
  reader_buff[count++] = ch;
  if (count == maxtoken-1)
    read_error_long(); /* fall through to fix up zero character */
  else
    goto more;
 done:
  reader_buff[count] = 0;
  return make_string(reader_buff);}


static object read_pound_sign() {
  int ch, count, i;
  rcatchpoint *thisrparen;
  boolean saved_ri;
  object arg1, arg2, items, vec;
  
  ch = nextchar(notoken);
  switch(ch){
  case '.': return eval(read_aux());
  case '\'': return list2(function,read_aux());
  case 'b':
  case 'B': return read_aux_with_base(2);
  case 'o':
  case 'O': return read_aux_with_base(8);
  case 'x':
  case 'X': return read_aux_with_base(16);
  case '(':
    reader_internal = true;
    items = nilobj;
    count = 0;
    reader_internal = true;
    goto startloop;
  loop:
    restore_rcatchpoint(thisrparen);
  startloop:
    rparen_catch = thisrparen = get_freercatchpoint();
    if (0 != setjmp(thisrparen->env))
      {restore_rcatchpoint(thisrparen);
      vec = make_vector(count);
      for (i=0; i<count; i++)
	{vector(vec)->arr[count-i-1] = first(items);
	items = rest(items);}
      return vec;}
    items = cons(read_aux(),items);
    ++count;
    goto loop;
  case '+': arg1 = read_aux();
    arg2 = read_aux();
    return read_decide_on_conditional(arg1, arg2);
  case '-': arg1 = read_aux();
    arg2 = read_aux();
    return read_decide_on_conditional(list2(notsymbol,arg1), arg2);
  case '|': saved_ri = reader_internal;
    reader_internal = true;
  nogood:
    while ('|' != nextchar(notoken));
  needdone:
    ch = nextchar(notoken);
    if (ch != '#')
      {if (ch == '|')
	goto needdone;
      else
	goto nogood;}
    reader_internal = saved_ri;
    return read_aux();
  default: read_error("Undefined # reader", ch); return nilobj;
  }}


static object read_aux_with_base (int base) {
  int savedbase;
  object ret;
	
  savedbase = fixnum(symbol(readbase)->value)->value;
  fixnum(symbol(readbase)->value)->value = base;
  ret = read_aux();
  fixnum(symbol(readbase)->value)->value = savedbase;
  return ret;}
	

static object read_decide_on_conditional (object testexp, object actualexp) {
  return read_conditional_truep(testexp)?actualexp:read_aux();}


static boolean read_conditional_truep (object testexp) {
  object lst;
	
  if (symbolp(testexp))
    return nilobj!=memq(testexp, features);
  else if (notpairp(testexp))
    error1("Malformed reader conditional expression: ~S", testexp);
  else if (andsymbol == car(testexp))
    {lst = testexp;
    while (nilobj!=(lst=rest(lst)))
      if (!read_conditional_truep(first(lst))) return false;
    return true;}
  else if (orsymbol == car(testexp))
    {lst = testexp;
    while (nilobj!=(lst=rest(lst)))
      if (read_conditional_truep(first(lst))) return true;
    return false;}
  else if (notsymbol == car(testexp))
    return !read_conditional_truep(second(testexp));
  else
    error1("Malformed reader conditional expression: ~S", testexp);}
  

static object read_is_number_p(char *str) {
  int i, base, count, lastindex, pointplace, eplace, digmax, alphamax;
  char *st, ch;
  double fnum; 
  boolean seenpoint = false;
  boolean seene = false;
  boolean digitseen = false;
  int num = 0;
  int sign = 1;
  int expsign = 0;
  int newdig = 0;
  int exponent = 0;
  
  base = fixnum(symbol(readbase)->value)->value;
  st = str;
  count=0;
  while (*st++) ++count;
  lastindex = count-1;
  for (i=0;i<count;i++)
    {ch=str[i];
    if ('.'==ch)
      {if (seenpoint || seene) return nilobj;
      seenpoint = true;
      pointplace = i;}
    if (((base == 10) || seenpoint)
	&&((ch == 'e') || (ch == 'E')))
      {if (seene || (i==0) || ((i+1)==count)) return nilobj;
      seene = true;
      eplace = i;}
    if (seene && ((ch == '+') || (ch == '-')))
      {if (! ((i==(1+eplace)) && ((i+1)<count))) return nilobj;
      expsign = (ch=='+')?1:-1;}}
  if (seene)
    if (count>eplace+(expsign!=0?1:0)+3) return nilobj;
  if (seenpoint) base = 10;
  digmax = '0' + -1 + ((10<base)?10:base);
  alphamax = 'A'+base-11;
  for (i=0;i<count;i++)
    {ch = str[i];
    ch = (('a'<=ch)&&(ch<='z'))?'A'+ch-'a':ch;
    if ((ch=='+') || (ch=='-'))
      {if (! ((i==0) || (expsign!=0))) return nilobj;
      if ((i==0) && (ch=='-')) sign = -1;}
    else if (seene && (i>=eplace))
      {if (i!=eplace)
	{if (('0'<=ch) && (ch<='9'))
	  exponent = 10*exponent + (ch-'0');
	else
	  return nilobj;}}
    else if ((('0'<=ch) && (ch<=digmax) && (newdig=ch-'0', true))
	     || (('A'<=ch) && (ch<=alphamax) && (newdig=10+ch-'A', true)))
      {/* removed decplaces and trailingzeroes processing here */
	num = newdig + base * num;
	digitseen = true;}
    else if (ch!='.') return nilobj;}
  if (seene || (seenpoint && (pointplace != lastindex)))
    /* we have a good floating point number! */
    return (1==sscanf(str, "%lf", &fnum))?make_flonum(fnum):nilobj;
  else
    return digitseen?make_fixnum(sign*num):nilobj;}


static object read_backquote() {
  int saved_bq;
  object ret;
  saved_bq = backquote_level++;
  ret = read_simplify(read_process_backquote(read_aux()));
  backquote_level = saved_bq;
  return ret;}


static object read_comma() {
  int saved_bq, ch;
  object ret;
  
  if (backquote_level <= 0)
    error0("Comma not inside backquote");
  else
    {ch=nextchar(notoken);
    saved_bq = backquote_level--;
    if (ch=='@')
      ret = cons(l__splice, read_aux());
    else if (ch=='.')
      ret = cons(l__dot, read_aux());
    else
      {backchar(ch);
      ret = cons(l__comma, read_aux());}
    backquote_level = saved_bq;
    return ret;}}
	  

static object read_process_backquote (object form) {
  object el, res;
  if (notpairp(form))
    return list2(quote,form);
  else if (l__comma==first(form))
    return rest(form);
  else if (l__splice==first(form))
    error1(",@~S after `", rest(form));
  else if (l__dot==first(form))
    error1(",.~S after `", rest(form));
  else
    {el=form;
    res=nilobj;
    while (pairp(el))
      {if (l__comma==first(el))
	return cons(l__append, nconc(nreverse(res), list1(rest(el))));
      if (l__splice==first(el))
	error1("Dotted ,@~S", rest(el));
      if (l__dot==first(el))
	error1("Dotted ,.~S", rest(el));
      res = cons(read_bracket(first(el)), res);
      el = rest(el);}
    return cons(l__append, nconc(nreverse(res), list1(list2(quote, el))));}}


static object read_bracket (object form) {
  if (notpairp(form))
    return list2(l__list, list2(quote, form));
  else if (first(form)==l__comma)
    return list2(l__list, rest(form));
  else if (first(form)==l__splice)
    return rest(form);
  else if (first(form)==l__dot)
    return list2(l__clobber, rest(form));
  else
    return list2(l__list, read_process_backquote(form));}


static object read_maptree_simplify (object x) {
  object a, d;
  if (notpairp(x)) return read_simplify(x);
  a = read_simplify(car(x));
  d = read_simplify(cdr(x));
  return ((a==car(x))&&(d==cdr(x)))?x:cons(a,d);}

	    
static object read_simplify (object x) {
  object newx, result, head, firstarg, args, items, revnew, fitem;
  boolean res;
	
  if (notpairp(x)) return x;
  newx = ((quote==first(x))&&(notpairp(second(x))))?x:read_maptree_simplify(x);
  if (l__append!=first(newx)) return newx;
  args = nreverse(rest(newx));
  result = nilobj;
 loop:
  if (nullp(args)) return result;
  firstarg = first(args);
  if (notpairp(firstarg))
    result = read_attach_append(l__append, firstarg, result);
  else
    {head = first(firstarg);
    res = ((head==l__list)||(head==l__liststar));
    if (res)
      for (items=rest(firstarg);pairp(items);items=rest(items))
	{fitem = first(items);
	if ((pairp(fitem))
	    &&((l__splice==first(fitem))||(l__dot==first(fitem))))
	  {res=false;
	  break;}}
    if (res)
      {if (head==l__list)
	result = read_attach_conses(rest(firstarg), result);
      else
	{revnew=nreverse(rest(firstarg));
	result = read_attach_conses(nreverse(rest(revnew)),
				    read_attach_append(l__append,
						       first(revnew),
						       result));}}
    else if ((quote==head) && (pairp(second(firstarg)))
	     &&(!(fitem=first(second(firstarg)),
		  ((l__comma==fitem)||(l__splice==fitem)||(l__dot==fitem))))
	     &&(nullp(rest(rest(firstarg)))))
      result = read_attach_conses(list1(list2(quote,first(second(firstarg)))),
				  result);
    else if (head==l__clobber)
      result = read_attach_append(l__nconc,rest(firstarg),result);
    else
      result = read_attach_append(l__append,firstarg,result);}
  args = rest(args);
  goto loop;}
	  

static object read_attach_append (object op, object item, object result) {
  if (((nullp(item))||((pairp(item))&&(quote==first(item))))
      &&((nullp(result))||((pairp(item))&&(quote==first(result)))))
    return list2(quote,append(second(item),second(result)));
  else if ((nullp(result))
	   ||((pairp(result))
	      &&(quote==first(result))
	      &&(nullp(second(result)))))
    {if ((pairp(item))&&((l__splice==first(item))||(l__dot==first(item))))
      return list2(op,item);
    else
      return item;}
  else if ((pairp(result))&&(op==first(result)))
    return cons(op,cons(item,rest(result)));
  else
    return list3(op,item,result);}


static object read_attach_conses (object items, object result) {
  object itemlist, item;
  boolean flag = true;	
	
  for (itemlist=items;pairp(itemlist);itemlist=rest(itemlist))
    {item=first(itemlist);
    if (! ((nullp(item))||((pairp(item))&&(quote==first(item)))))
      {flag = false;
      break;}}
  if (flag && ((nullp(result))||((pairp(result))&&(quote==first(result)))))
    return list2(quote,append(mapsecond(items),second(result)));
  else if ((nullp(result))||(bequal(result,(list2(quote,nilobj)))))
    return cons(l__list,items);
  else if ((pairp(result))
	   &&((l__list==first(result))||(l__liststar==first(result))))
    return cons(first(result),append(items,rest(result)));
  else
    return cons(l__liststar,append(items,list1(result)));}
	

static object mapsecond (object lst) {
  if (nullp(lst))
    return nilobj;
  else
    return cons(second(first(lst)),mapsecond(rest(lst)));}


/* low level interfaces to io system */

static int nextchar (int tokenp) {
  int index, lastchar;
  char *str;
  FILE *openf;
	
  lastchar = instream->lastchar;
  str = instream->filename;
  openf = instream->openfile;
	
  if (-1!=lastchar)
    {instream->lastchar = -1;
    return lastchar;}
  else if (NULL==openf)
    {index=++instream->current;
    if (index>=instream->limit)
      {if (reader_internal)
	{unexpected_end_error(str);
	return 0;}
      else if (tokenp==token)
	return -1;
      else longjmp(eof_catch->env, 1);
      }
    else
      return str[index];}
  else if (openf==(void *)-1)
    {for(lastchar=getchar();lastchar==-1;lastchar=getchar())
      background(fell_notbusy);
    return lastchar;}
  else
    {lastchar = fgetc(instream->openfile);
    if (lastchar != -1)
      return lastchar;
    else if (reader_internal)
      {unexpected_end_error(instream->filename);
      return 0;}
    else if (tokenp==token)
      return -1;
    else longjmp(eof_catch->env, 1);
    }}


static void backchar (char ch) {
  if (-1!=instream->lastchar) read_error("Backing up two characters", ch);
  instream->lastchar=ch;
  return;}


static object fellopen (object filename, object restvars) {
  object direct, super;
  pmethod pmeth;
  lstream retl;
  boolean inputp = true;
  boolean appendp = false;
	
  if (!stringp(filename))
    error1("Need a string for a filename: ~S", filename);
  checkkeywordargs(restvars, 4, openkeywords);
  direct = memq(directkey,restvars);
  if (direct!=nilobj)
    {if (second(direct)==inputkey)
      inputp = true;
    else if (second(direct)==outputkey)
      inputp = false;
    else error1("Unknown file direction: ~S", second(direct));}
  retl = malloc(sizeof(streamstruct));
  retl->prmeth = NULL;
  if (!inputp)
    {super = memq(ifexistskey,restvars);
    if (super!=nilobj)
      {if (second(super)==appendkey)
	appendp = true;
      else if (second(super)==supersedekey)
	appendp = false;
      else error1("Unknown :if-exists keyword: ~S", second(super));}
    pmeth = new_pmethod();
    pmeth->stream = retl;
    retl->prmeth = pmeth;}
  retl->lastchar = -1;
  retl->filename = string(filename)->chars;
  if (NULL ==
      (retl->openfile = fopen(string(filename)->chars,
			      inputp?"r":appendp?"a":"w")))
    error1("Unable to open file ~S", filename);
  retl->openstatus = true;
  /* all in static space -- no gc issues */
  return cons(filekey,cons(inputp?inputkey:outputkey,(void *)retl));}


static int felllength (object lst) {
  int count = 0;
  
  while (lst != nilobj)
    {count++;
    lst = rest(lst);}
  return count;}
	

static void checkkeywordargs (object args, int arglength, object validkeys) {
  int len;
  object argl;
	
  len = felllength(args);
  if (0!=(1&len)) error1("Odd number of keyword arguments: ~S", args);
  if (len>arglength) error1("Too many keyword arguments: ~S", args);
	
  argl = args;
  while (argl != nilobj)
    {if (nilobj==memq(first(argl), validkeys))
      error2("Invalid keyword ~S in arglist: ~S", first(args), args);
    argl = rest(rest(argl));}
  return;}


static object fellclose (object stream) {
  if (pairp(stream)
      &&(first(stream) == filekey)
      &&true == ((lstream)rest(rest(stream)))->openstatus)
    {fclose(((lstream)rest(rest(stream)))->openfile);
    ((lstream)rest(rest(stream)))->openstatus = false;
    return tsymbol;}
  else
    error1("Not a closeable file: ~S", stream);}


/* various error handlers */

static void read_error (char *mess, char ch) {
  printf("\n\n**************************************");
  printf("\n%s: %d -- %c\n\n", mess, ch, ch);
  return;}


static void read_error_long () {
  printf("\n\n**************************************");
  printf("\nToken too long: %s\n\n", &reader_buff);
  return;}

	
static void read_lerror (char ch, int attr) {
  printf("\n\n**************************************");
  printf("\n%s escape characters are not handled by FELL: %d -- %c\n\n",
	 (singleescapep(attr))?"Single":"Multiple",
	 ch,
	 ch);
  return;}


static void read_general_error (char *str) {
  printf("\n\n**************************************");
  printf("\n%s error in READ", str);
  return;}


static void unexpected_end_error (char *str) {
  printf("\n\n**************************************");
  printf("\nUnexpected end of file in: %s", str);
  exit(-1);
  return;}


static object end_of_file_error (lstream stream,
				 object eof_error_p,
				 object eof_return) {
  if (nilobj==eof_error_p)
    return eof_return;
  else
    {printf("\n\n**************************************");
    printf("\nEnd of file: %d\n\n", stream);
    return nilobj;}}


/*****************************************************************************
******************************************************************************
******************************************************************************

*** the primops

******************************************************************************
******************************************************************************
*****************************************************************************/


static object memq (object item, object lst) {
  object down;

  check_listnil(lst);	
  down = lst;
 loop:
  if (nullp(down)) return nilobj;
  if (notpairp(down)) error1("Malformed list in MEMQ: ~S", lst);
  if (item == car(down))
    return down;
  else
    {down = cdr(down);
    goto loop;}}


static object assq (object item, object lst) {
  object down;

  check_listnil(lst);	
  down = lst;
 loop:
  if (nullp(down)) return nilobj;
  if (notpairp(down)) error1("Malformed list in ASSQ: ~S", lst);
  if (item == car(car(down)))
    return car(down);
  else
    {down = cdr(down);
    goto loop;}}


static object append (object u, object v) {
  object tmp;
  check_listnil(u);	
  if (u==nilobj)
    return v;
  else {
    savereg(first(u));
    tmp = append(rest(u), v);
    restorereg(u);
    return cons(u, tmp);}}


static object nreverse (object lst) {
  object sofar, rem, tmp;

  check_listnil(lst);	
  sofar = nilobj;
  rem = lst;
  while (nilobj != rem)
    {tmp = rest(rem);
    rest(rem) = sofar;
    sofar = rem;
    rem = tmp;}
  return sofar;}


static object reverse (object lst) {
  object res;
  check_listnil(lst);
  savereg(lst);
  savereg(nilobj); /* res */
  while (nilobj!=referreg(1)) {
    referreg(0) = cons(first(referreg(1)),referreg(0));
    referreg(1) = cdr(referreg(1));}
  restorereg(res);
  restorereg(lst); /* pop stack */
  return res;}


static object nconc (object lst1, object lst2) {
  object cell;

  check_listnil(lst1);	
  if (nullp(lst1))
    return lst2;
  else
    {cell = lst1;
    while (pairp(rest(cell)))
      cell = rest(cell);
    rest(cell) = lst2;
    return lst1;}}


static object nthproc (object index, object lst) {
  int i;

  check_fixnum(index);
  i = fixnum(index)->value;

 loop:
  if (lst==nilobj)
    return nilobj;
  check_cons(lst);
  if (i<=0) return first(lst);
  i--;
  lst = rest(lst);
  goto loop;}

static object setproc (object symb, object val) {
  if ((!symbolp(symb)) || (nilobj!=is_constant_p(symb)))
    error1("Can only set non-constant symbols, not: ~S", symb);
  if (nilobj==is_special_p(symb)) {
    warn1("Setting non-special ~S -- making it special", symb);
    declare_special(symb);}
  symbol(symb)->value = val;
  return val;}

static object atomproc (object arg) {return pairp(arg)?nilobj:tsymbol;}
static object conspproc (object arg) {return pairp(arg)?tsymbol:nilobj;}
static object nullproc (object arg) {return nullp(arg)?tsymbol:nilobj;}
static object symbolpproc (object arg) {return symbolp(arg)?tsymbol:nilobj;}
static object vectorpproc (object arg) {return vectorp(arg)?tsymbol:nilobj;}
static object stringpproc (object arg) {return stringp(arg)?tsymbol:nilobj;}
static object procedurepproc (object arg) {return procedurep(arg)?tsymbol:nilobj;}
static object structurepproc (object arg) {return structurep(arg)?tsymbol:nilobj;}
static object fixnumpproc (object arg) {return fixnump(arg)?tsymbol:nilobj;}
static object flonumpproc (object arg) {return flonump(arg)?tsymbol:nilobj;}
static object numberpproc (object arg) {
  return (fixnump(arg)||flonump(arg))?tsymbol:nilobj;}
static object eqproc (object arg1, object arg2) {
  return (arg1==arg2)?tsymbol:nilobj;}

static object numeqproc (object item1, object item2) {
  check_number(item1);
  check_number(item2);
  if (fixnump(item1))
    if (fixnump(item2))
      return (fixnum(item1)->value == fixnum(item2)->value)?tsymbol:nilobj;
    else
      return (((double)fixnum(item1)->value) == flonum(item2)->value)?tsymbol:nilobj;
  else
    if (fixnump(item2))
      return (flonum(item1)->value == (double)fixnum(item2)->value)?tsymbol:nilobj;
    else
      return (flonum(item1)->value == flonum(item2)->value)?tsymbol:nilobj;}

static object numneproc (object item1, object item2) {
  check_number(item1);
  check_number(item2);
  if (fixnump(item1))
    if (fixnump(item2))
      return (fixnum(item1)->value == fixnum(item2)->value)?nilobj:tsymbol;
    else
      return (((double)fixnum(item1)->value) == flonum(item2)->value)?nilobj:tsymbol;
  else
    if (fixnump(item2))
      return (flonum(item1)->value == (double)fixnum(item2)->value)?nilobj:tsymbol;
    else
      return (flonum(item1)->value == flonum(item2)->value)?nilobj:tsymbol;}

static object numltproc (object item1, object item2) {
  check_number(item1);
  check_number(item2);
  if (fixnump(item1))
    if (fixnump(item2))
      return (fixnum(item1)->value < fixnum(item2)->value)?tsymbol:nilobj;
    else
      return (((double)fixnum(item1)->value) < flonum(item2)->value)?tsymbol:nilobj;
  else
    if (fixnump(item2))
      return (flonum(item1)->value < (double)fixnum(item2)->value)?tsymbol:nilobj;
    else
      return (flonum(item1)->value < flonum(item2)->value)?tsymbol:nilobj;}

static object numleproc (object item1, object item2) {
  check_number(item1);
  check_number(item2);
  if (fixnump(item1))
    if (fixnump(item2))
      return (fixnum(item1)->value <= fixnum(item2)->value)?tsymbol:nilobj;
    else
      return (((double)fixnum(item1)->value) <= flonum(item2)->value)?tsymbol:nilobj;
  else
    if (fixnump(item2))
      return (flonum(item1)->value <= (double)fixnum(item2)->value)?tsymbol:nilobj;
    else
      return (flonum(item1)->value <= flonum(item2)->value)?tsymbol:nilobj;}

static object numgtproc (object item1, object item2) {
  check_number(item1);
  check_number(item2);
  if (fixnump(item1))
    if (fixnump(item2))
      return (fixnum(item1)->value > fixnum(item2)->value)?tsymbol:nilobj;
    else
      return (((double)fixnum(item1)->value) > flonum(item2)->value)?tsymbol:nilobj;
  else
    if (fixnump(item2))
      return (flonum(item1)->value > (double)fixnum(item2)->value)?tsymbol:nilobj;
    else
      return (flonum(item1)->value > flonum(item2)->value)?tsymbol:nilobj;}

static object numgeproc (object item1, object item2) {
  check_number(item1);
  check_number(item2);
  if (fixnump(item1))
    if (fixnump(item2))
      return (fixnum(item1)->value >= fixnum(item2)->value)?tsymbol:nilobj;
    else
      return (((double)fixnum(item1)->value) >= flonum(item2)->value)?tsymbol:nilobj;
  else
    if (fixnump(item2))
      return (flonum(item1)->value >= (double)fixnum(item2)->value)?tsymbol:nilobj;
    else
      return (flonum(item1)->value >= flonum(item2)->value)?tsymbol:nilobj;}


static boolean bequal (object item1, object item2) {
  if (item1==item2) return true;
  else if (nullp(item1)) return nullp(item2);
  else if (fixnump(item1))
    return ((fixnump(item2))&&fixnum(item1)->value==fixnum(item2)->value);
  else if (flonump(item1))
    return ((flonump(item2))&&flonum(item1)->value==flonum(item2)->value);
  else if (pairp(item1))
    {if (notpairp(item2))
      return false;
    else
      return ((bequal(car(item1),car(item2)))&&bequal(cdr(item1),cdr(item2)));}
  else if (string(item1))
    {if (stringp(item2))
      return !different_strings(string(item1)->chars, item2);
    else
      return false;}
  else return false;}


static object equalproc (object item1, object item2) {
  return bequal(item1,item2)?tsymbol:nilobj;}
	
static object carproc (object x) {
  check_consnil(x, "car"); return pairp(x)?car(x):nilobj;}
static object cdrproc (object x) {
  check_consnil(x, "cdr"); return pairp(x)?cdr(x):nilobj;}
static object secondproc (object x) {
  check_consnil(x, "second");
  return (pairp(x)&&pairp(cdr(x)))?car(cdr(x)):nilobj;}
static object thirdproc (object x) {
  check_consnil(x, "third");
  return (pairp(x)&&pairp(cdr(x))&&pairp(cdr(cdr(x))))?car(cdr(cdr(x))):nilobj;}
static object listproc (object restarg) {return restarg;}

static object liststarproc (object fst, object restargs) {
  object tmp;
  if (nullp(restargs))
    return fst;
  else {
    savereg(fst);
    tmp = liststarproc(first(restargs), rest(restargs));
    restorereg(fst);
    return cons(fst, tmp);}}


static object fell_boundp (object sym) {
  if (!symbolp(sym)) error1("In boundp ~S is not a symbol", sym);
  if (unbound_value==symbol(sym)->value)
    return nilobj;
  else
    return tsymbol;}
	

static object keywordp (object sym) {
  return ((symbolp(sym))
	  &&(':'==string(symbol(sym)->name)->chars[0]))?tsymbol:nilobj;}
	  

static object adder (object restargs) {
  object test, arg;
  boolean allfix = true;
  int sumi = 0;
  double sumf = 0.0; 
	
  test = restargs;
  while (test!=nilobj)
    {arg = first(test);
    if (fixnump(arg)) {}
    else if flonump(arg) allfix = false;
    else
      error_needallnumbers(restargs);
    test = rest(test);}
  if (allfix)
    {test = restargs;
    while (test!=nilobj)
      {sumi = sumi + fixnum(first(test))->value;
      test = rest(test);}
    return make_fixnum(sumi);}
  else  
    {test = restargs;
    while (test!=nilobj)
      {arg = first(test);
      sumf = sumf +
	((flonump(arg))?flonum(arg)->value:(double)(fixnum(arg)->value));
      test = rest(test);}
    return make_flonum(sumf);}}
	

static object subber (object firstarg, object restargs) {
  object test, arg;
  double fstf;
  boolean allfix = true;
  int sumi = 0;
  double sumf = 0.0; 
	
  if (fixnump(firstarg)) {}
  else if (flonump(firstarg)) allfix = false;
  else error_needallnumbers(restargs);
	
  test = restargs;
  while (test!=nilobj)
    {arg = first(test);
    if (fixnump(arg)) {}
    else if (flonump(arg)) allfix = false;
    else
      error_needallnumbers(restargs);
    test = rest(test);}
  if (allfix)
    {test = restargs;
    while (test!=nilobj)
      {sumi = sumi + fixnum(first(test))->value;
      test = rest(test);}
    return make_fixnum((nilobj==restargs)?(0-fixnum(firstarg)->value)
		       :(fixnum(firstarg)->value)-sumi);}
  else  
    {test = restargs;
    while (test!=nilobj)
      {arg = first(test);
      sumf = sumf +
	((flonump(arg))?flonum(arg)->value:(double)(fixnum(arg)->value));
      test = rest(test);}
    fstf = (flonump(firstarg))?flonum(firstarg)->value:(double)(fixnum(firstarg)->value);
    return make_flonum((nilobj==restargs)?(0.0-fstf):fstf-sumf);}}
	  

static object timeser (object restargs) {
  object test, arg;
  boolean allfix = true;
  int prodi = 1;
  double prodf = 1.0; 
	
  test = restargs;
  while (test!=nilobj)
    {arg = first(test);
    if (fixnump(arg)) {}
    else if (flonump(arg))
      allfix = false;
    else
      error_needallnumbers(restargs);
    test = rest(test);}
  if (allfix)
    {test = restargs;
    while (test!=nilobj)
      {prodi = prodi * fixnum(first(test))->value;
      test = rest(test);}
    return make_fixnum(prodi);}
  else  
    {test = restargs;
    while (test!=nilobj)
      {arg = first(test);
      prodf = prodf *
	((flonump(arg))?flonum(arg)->value:(double)(fixnum(arg)->value));
      test = rest(test);}
    return make_flonum(prodf);}}
	
	  
static object divider (object firstarg, object restargs) {
  object test, arg;
  double fstf;
  int prodi = 1;
  double prodf = 1.0; 
	
  if (fixnump(firstarg)) {}
  else if (flonump(firstarg)) {}
  else error_needallnumbers(restargs);
	
  test = restargs;
  while (test!=nilobj)
    {arg = first(test);
    if (fixnump(arg)) {}
    else if (flonump(arg)) {}
    else
      error_needallnumbers(restargs);
    test = rest(test);}
  test = restargs;
  while (test!=nilobj)
    {arg = first(test);
    prodf = prodf *
      ((flonump(arg))?flonum(arg)->value:(double)(fixnum(arg)->value));
    test = rest(test);}
  fstf = (flonump(firstarg))?flonum(firstarg)->value
    :(double)(fixnum(firstarg)->value);
  fstf = (nilobj==restargs)?(1.0/fstf):fstf/prodf;
  return (fstf==(double)(int)fstf)?make_fixnum((int)fstf):make_flonum(fstf);}
	  

static object logander (object restargs) {
  object test, arg;
  int ander = -1;
	
  test = restargs;
  while (test!=nilobj)
    {arg = first(test);
    if (!fixnump(arg))
      error_needallfixnums(restargs);
    ander = ander & fixnum(arg)->value;
    test = rest(test);}
  return make_fixnum(ander);}
	

static object sqrtproc (object num) {
  double val;
  check_number(num);
  val = (fixnump(num))?((double)fixnum(num)->value):flonum(num)->value;
  if (val<0.0) error1("Argument to sqrt is negative: ~S", num);
  return make_flonum(sqrt(val));}


static object lshproc (object num, object shift) {
  int sval;
  check_fixnum(num);
  check_fixnum(shift);
  sval = fixnum(shift)->value;
  if (sval<0) error1("Shift should be non-negative: ~S", shift);
  return make_fixnum(sval>=32?0:(fixnum(num)->value)<<sval);}


static object floatproc (object num) {
  double val;
  check_number(num);
  return make_flonum((fixnump(num))
		     ?((double)fixnum(num)->value)
		     :flonum(num)->value); }



static object logiorer (object restargs) {
  object test, arg;
  int orer = 0;
	
  test = restargs;
  while (test!=nilobj)
    {arg = first(test);
    if (!fixnump(arg))
      error_needallfixnums(restargs);
    orer = orer | fixnum(arg)->value;
    test = rest(test);}
  return make_fixnum(orer);}
	

static object logxorer (object restargs) {
  object test, arg;
  int xorer = 0;
	
  test = restargs;
  while (test!=nilobj)
    {arg = first(test);
    if (!fixnump(arg))
      error_needallfixnums(restargs);
    xorer = xorer ^ fixnum(arg)->value;
    test = rest(test);}
  return make_fixnum(xorer);}
	

static object lognoter (object arg) {
  if (!fixnump(arg))
    error1("Argument to lognot must be a fixnum: ~S", arg);
  else
    return make_fixnum(~fixnum(arg)->value);}


static object svrefproc (object vect, object index) {
  check_vector_args(vect,index);
  return vector(vect)->arr[fixnum(index)->value];}
	
	
static object svsetproc (object vect, object index, object val) {
  check_vector_args(vect,index);
  vector(vect)->arr[fixnum(index)->value] = val;
  return val;}

	
static void check_vector_args (object vec, object index) {
  int len, ix;
  if (!(vectorp(vec)))
    error2("Expected a vector in a vector reference: ~S[~S]", vec, index);
  if (!(fixnump(index)))
    error2("Expected fixnum index for vector reference:  ~S[~S]", vec, index);
  len = vector(vec)->length;
  ix = fixnum(index)->value;
  if ((ix < 0) || (len <= ix))
    error2("Index ~S is out of bounds for vector: ~S", index, vec);
  return;}

	
static object mvecproc (object size, object restargs) {
  int len, i;
  object init, ret;
	
  if (!((fixnump(size)) && (fixnum(size)->value > -1)))
    error1("To make a vector must specify fixnum elements at least zero, not ~S",
	   size);
  init = nilobj;
	
  if (!(nilobj == restargs))
    {if (!((initelemkey==first(restargs))
	   && (pairp(rest(restargs)))
	   && (nullp(rest(rest(restargs))))))
      error1("Malformed arglist to make vector: ~S", restargs);
    init = second(restargs);}
  len = fixnum(size)->value;
  savereg(init);
  ret = make_vector(len);
  restorereg(init);
  if (nilobj != init)
    for(i=0;i<len;i++) vector(ret)->arr[i]=init;
  return ret;}


static object lengthproc (object obj) {
  int len;
  object cop;
  switch(gettype(obj)){
  case type_null: return zerobj;
  case type_string: return make_fixnum(string(obj)->length);
  case type_vector: return make_fixnum(vector(obj)->length);
  case type_cons:
    len = 1;
    for (cop=rest(obj);pairp(cop);cop=rest(cop)) len++;
    if (!nullp(cop))
      error1("Object ~S terminates unlike a list--no length", obj);
    return make_fixnum(len);
  default: error1("~S is not a type which has a length", obj);}}


static object gensymstring;
static object gensymlist;


static object gensymproc () {	
  object count;

  count = symbol(gensymcounter)->value;
  symbol(gensymcounter)->value = make_fixnum(1+fixnum(count)->value);
  first(gensymlist) = count;
  return intern(string(format(nilobj,gensymstring,gensymlist))->chars);}


static object mapcarproc (object proc, object lst1, object restargs) {
  object tmp;
  sstack savedstack;
	
  savereg(proc);
  restargs = cons(lst1, restargs);
  savereg(restargs);
  savereg(nilobj);
  savereg(nilobj);
  savereg(nilobj);
  /* stack is arglist, result, scanlist, restargs, proc */
  savestack(savedstack);
  while (true) {
    referreg(0) = nilobj;
    for (referreg(2)=referreg(3);
	 referreg(2)!=nilobj;
	 referreg(2)=cdr(referreg(2))) {
      if (nilobj==first(referreg(2)))
	goto done;
      else {
	referreg(0) = cons(first(first(referreg(2))), referreg(0));
	first(referreg(2)) = rest(first(referreg(2)));}}
    tmp = applyproc(referreg(4),nreverse(referreg(0)));
    restorestack(savedstack);
    referreg(1) = cons(tmp, referreg(1));}
 done:	
  tmp = referreg(1);
  regsp -= 5;
  return nreverse(tmp);}

	
static object check_ltype_proc (object obj, object str) {
  int type;
  type = gettype(obj);
  if ((type_last <= type)&&(next_ltype > type)&&(stringp(str))
      &&!(different_strings(string(car(vector(ltypes)->arr[type-type_last]))->chars,
			    str)))
    return tsymbol;
  else
    return nilobj;}
  

/* error handlers */

static object errorstring;
static object debugstring1;
static object debugstring2;
static object debugstring3;
static object debugstring4;
static object debugstring5;
static object debugstring6;
static object debugstring7;
static object debugstring8;
static object debugstring9;
static object debugstring10;
static object debugstring11;
static object debugstring12;
static object debugstring13;
static object debugstring14;
static object debugstring15;
static object debugstring16;
static object debugstring17;
static object debuglist5;
static object debuglist6;
static object debuga;
static object debugb;
static object debugf;
static object debugh;
static object debugm;
static object debugq;
static object debugsb;
static object debugsln;
static object debugslv;
static object debugv;
static object inspectstring0;
static object inspectstring1;
static object inspectstring2;
static object inspectstring3;
static object inspectstring4;
static object inspectstring5;
static object inspectstring6;
static object inspectstring7;
static object inspecth;
static object inspectq;
static object inspectu;
static object inspectr;
static object inspectsb;
static object inspectsln;
static object inspectslv;

static void init_error_handlers () {
  gensymstring = make_string("g~4,'0D");
  gensymlist = cons(nilobj,nilobj);
  warnstring = make_string("~&;;; Warning: ");
  errorstring = make_string("~&;;; ERROR: ");
  debugstring1 = make_string("~&debug? ");
  debugstring2 = make_string("~&;;; Debug command must be a keyword (try :h)");
  debugstring3 = make_string("~&Debugger commands:~%  :a~10Tabort to top level~%  :b~10Tbacktrace~%  :f n~10Tframe n~%  :h~10Thelp~%  :m~10Tredisplay message~%  :q~10Tquit from FELL~%  :sb n~10Tset print base~%  :sln n~10Tset print length~%  :slv n~10Tset print level~%  :v~10Tvariables in frame 0");
  debugstring4 = make_string("~&;;; Invalid debugger command (try :h)");
  debugstring5 = make_string("[~D]~S <- ");
  debugstring6 = make_string("toplevel");
  debugstring7 = make_string("~&   ~S:~12T~S");
  debugstring8 = make_string("~&;;; Expected a numeric argument: ~S");
  debugstring9 = make_string("~&[~D]~S");
  debugstring10 = make_string("~& Block {~S}");
  debugstring11 = make_string("~& Lexical let bindings");
  debugstring12 = make_string("~&   ~S~20Trebind to ~S");
  debugstring13 = make_string("~& Saved special values");
  debugstring14 = make_string("~& Tags {");
  debugstring15 = make_string("~S, ");
  debugstring16 = make_string("~S}");
  debugstring17 = make_string("~& eval form: ~S");
  debuglist5 = list2(make_fixnum(fixnumoffset+fixnumoffset), nilobj);
  debuglist6 = list2(nilobj,nilobj);
  debuga = intern(":a");
  debugb = intern(":b");
  debugf = intern(":f");
  inspecth = debugh = intern(":h");
  debugm = intern(":m");
  inspectq = debugq = intern(":q");
  inspectr = intern(":r");
  inspectu = intern(":u");
  inspectsb = debugsb = intern(":sb");
  inspectsln = debugsln = intern(":sln");
  inspectslv = debugslv = intern(":slv");
  debugv = intern(":v");

  inspectstring0 = make_string("~&   ~S");
  inspectstring1 = make_string("~&inspect? ");
  inspectstring2 = make_string("~&;;; Invalid inspector command (try :h)");
  inspectstring3 = make_string("~&>>Inspector commands:~%  :h~10Thelp~%  :q~10Tquit inspector~%  :r~10Tredisplay~%  :sb n~10Tset print base~%  :sln n~10Tset print length~%  :slv n~10Tset print level~%  :u~10Tup to previous item");
  inspectstring4 = make_string(" ~S");
  inspectstring5 = make_string("~%  [~s..~s]~10Tselect component");
  inspectstring6 = make_string("~&;;; Object is not an indexable type");
  inspectstring7 = make_string("~&;;; Index out of bounds");
  return;}


static void error0 (char *errorstr) {
  object str;
  format(tsymbol, errorstring, nilobj);
  enter_debugger();
  format(tsymbol, (str=make_string(errorstr)), nilobj);
  terpri(tsymbol);
  error_recover(str, nilobj);}


static void error1 (char *errorstr, object item) {
  object str, lst;
  format(tsymbol, errorstring, nilobj);
  enter_debugger();
  lst=cons(item, nilobj);
  savereg(lst);
  str=make_string(errorstr);
  restorereg(lst);
  format(tsymbol, str, lst);
  terpri(tsymbol);
  error_recover(str, lst);}


static void error2 (char *errorstr, object item1, object item2) {
  object str, lst;
  format(tsymbol, errorstring, nilobj);
  enter_debugger();
  lst=cons(item2, nilobj);
  lst=cons(item1, lst);
  savereg(lst);
  str=make_string(errorstr);
  restorereg(lst);
  format(tsymbol, str, lst);
  terpri(tsymbol);
  error_recover(str, lst);}


static object errorproc (object string, object restargs) {
  format(tsymbol, errorstring, nilobj);
  enter_debugger();
  format(tsymbol, string, restargs);
  terpri(tsymbol);
  error_recover(string, restargs);}
	

static void warn1 (char *warnstr, object item) {
  object str, lst;
  format(tsymbol, warnstring, nilobj);
  lst=cons(item, nilobj);
  savereg(lst);
  str=make_string(warnstr);
  restorereg(lst);
  format(tsymbol, str, lst);
  terpri(tsymbol);
  return;}


static void warn2 (char *warnstr, object item1, object item2) {
  object str, lst;
  format(tsymbol, warnstring, nilobj);
  lst=cons(item2, nilobj);
  lst=cons(item1, lst);
  savereg(lst);
  str=make_string(warnstr);
  restorereg(lst);
  format(tsymbol, str, lst);
  terpri(tsymbol);
  return;}


static object warnproc (object string, object restargs) {
  format(tsymbol, warnstring, nilobj);
  format(tsymbol, string, restargs);
  terpri(tsymbol);
  return nilobj;}


static void error_needallnumbers (object restargs) {
  error1("Need all numbers for arithmetic, not ~S", restargs);
  return;}


static void error_needallfixnums (object restargs) {
  error1("Need all fixnums for logic, not ~S", restargs);
  return;}


static void check_listnil (object lst) {
  object shimmy;
  shimmy = lst;
  while (nilobj!=shimmy)
    {if (notpairp(shimmy))
      error1("Malformed list ~S", lst);
    shimmy = rest(shimmy);}
  return;}
	

static void check_cons (object conshuh) {
  if (pairp(conshuh))
    return;
  else error1("Not a cons ~S", conshuh);}


static void check_consnil (object conshuh, char *str) {
  if (pairp(conshuh)||nullp(conshuh))
    return;
  else error2("Not a cons/nil ~S in ~A", conshuh, make_string(str));}


static void check_number (object num) {
  if ((fixnump(num)) || (flonump(num)))
    return;
  else error1("Not a number ~S", num);}
		

static void check_fixnum (object num) {
  if (fixnump(num))
    return;
  else error1("Not a fixnum ~S", num);}


/* this is not GC safe, so we force a GC first and hope for
   the best */

static void error_recover (object str, object lst) {
  object command, scanstack;
  int index, spp;
	
  /*
  savereg(str);
  savereg(lst);
  gc();
  restorereg(lst);
  restorereg(str);
  */

 loop:
  format(tsymbol, debugstring1, nilobj);
  command = readproc(tsymbol, nilobj, nilobj);
  tprintmethod->column = 0;
  if (nilobj==keywordp(command))
    {format(tsymbol, debugstring2, nilobj);
    goto loop;}
  else if (debugh == command)
    {format(tsymbol, debugstring3, nilobj);
    goto loop;}
  else if (debugq == command) 
    {exit_debugger();
    simthrow(warnstring,nilobj);}
  else if (debuga == command)
    {exit_debugger();
    simthrow(prompt,nilobj);}
  else if (debugm == command)
    {format(tsymbol, errorstring, nilobj);
    format(tsymbol, str, lst);
    goto loop;}
  else if (debugv == command)
    {display_frame(0);
    goto loop;}
  else if (debugf == command)
    {command = readproc(tsymbol, nilobj, nilobj);
    if (!fixnump(command))
      {format(tsymbol, debugstring8, cons(command, nilobj));
      goto loop;}
    display_frame(fixnum(command)->value);
    goto loop;}
  else if (debugb == command) {
    spp = sp;
    for (index = 0; true; index++) {
      while (((scanstack=stack[spp])!=nilobj)
	     &&(first(first(scanstack))!=framesymbol))
	spp--;
      if (scanstack==nilobj) break;
      fixnum(first(debuglist5))->value = index;
      second(debuglist5) = frameproc(first(scanstack));
      format(tsymbol,debugstring5,debuglist5);
      spp--;}
    format(tsymbol,debugstring6,nilobj);
    goto loop;}
  else if (debugsb == command) 
    {command = readproc(tsymbol, nilobj, nilobj);
    if (debug_valid_p(command, false, 2))
      {symbol(printbase)->value = command;
      symbol(debugprintbase)->value = command;}
    goto loop;}
  else if (debugsln == command)  
    {command = readproc(tsymbol, nilobj, nilobj);
    if (debug_valid_p(command, true, 1))
      {symbol(printlength)->value = command;
      symbol(debugprintlength)->value = command;}
    goto loop;}
  else if (debugslv == command)  
    {command = readproc(tsymbol, nilobj, nilobj);
    if (debug_valid_p(command, true, 1))
      {symbol(printlevel)->value = command;
      symbol(debugprintlevel)->value = command;}
    goto loop;}
  else
    {format(tsymbol, debugstring4, nilobj);
    goto loop;}}


static object saved_plevel, saved_plength, saved_pbase;


static void enter_debugger () {
  saved_plevel = symbol(printlevel)->value;
  saved_plength = symbol(printlength)->value;
  saved_pbase = symbol(printbase)->value;
  symbol(printlevel)->value = symbol(debugprintlevel)->value;
  symbol(printlength)->value = symbol(debugprintlength)->value;
  symbol(printbase)->value = symbol(debugprintbase)->value;
  return;}
	

static void exit_debugger () {
  symbol(printlevel)->value = saved_plevel;
  symbol(printlength)->value = saved_plength;
  symbol(printbase)->value = saved_pbase;
  return;}


static boolean debug_valid_p (object item, boolean nilok, int minnum) {
  if (fixnump(item))
    {if (fixnum(item)->value >= minnum)
      return true;
    else
      {printf("Fixnum argument must be at least %d\n", minnum);
      return false;}}
  else if (nilok)
    {if (item==nilobj)
      return true;
    else
      {printf("Argument must be () or fixnum at least %d\n", minnum);
      return false;}}
  else
    {printf("Argument must be a fixnum at least %d\n", minnum);
    return false;}}


static void display_frame (int indx) {
  object scanstack, item, type;
  int index, spp;
	
  index = indx;
  spp = sp;
  while (index!=0) {
    while (((scanstack=stack[spp])!=nilobj)
	   &&(first(first(scanstack))!=framesymbol))
      spp--;
    spp--;
    if (spp==0) goto overflow;
    index--;}
  scanstack=stack[spp];
  while (scanstack!=nilobj)
    {item = first(scanstack);
    type = first(item);
    if (type==framesymbol) {
      type = frameproc(item);
      for (item=rest(rest(item));item!=nilobj;item=rest(item))
	{first(debuglist6) = first(first(item));
	second(debuglist6) = rest(first(item));
	format(tsymbol,debugstring7,debuglist6);}
      fixnum(first(debuglist5))->value = indx;
      second(debuglist5) = type;
      format(tsymbol,debugstring9,debuglist5);}
    else if (type==letsymbol)
      {for (item=rest(item);item!=nilobj;item=rest(item))
	{first(debuglist6) = first(first(item));
	second(debuglist6) = rest(first(item));
	format(tsymbol,debugstring7,debuglist6);}
      format(tsymbol,debugstring11,nilobj);}
    else if (type==blocksymbol)
      {second(debuglist5) = rest(item);
      format(tsymbol,debugstring10,rest(debuglist5));}
    else if (type==progvsymbol)
      {for (item=rest(item);item!=nilobj;item=rest(item))
	{first(debuglist6) = first(first(item));
	second(debuglist6) = rest(first(item));
	format(tsymbol,debugstring7,debuglist6);}
      format(tsymbol,debugstring13,nilobj);}
    else if (type==evalsymbol)
      {second(debuglist5) = rest(item);
      format(tsymbol,debugstring17,rest(debuglist5));}
    else if (type==tagbodysymbol)
      {format(tsymbol,debugstring14,nilobj);
      for (item=rest(item);item!=nilobj;item=rest(item))
	{second(debuglist6) = first(first(item));
	format(tsymbol,
	       (nullp(rest(item)))?debugstring16:debugstring15,
	       rest(debuglist6));}}
    scanstack = rest(scanstack);}
  return;
 overflow:
  printf("Can not find frame indexed with: %d\n", indx);
  return;}



/*****************************************************************************
******************************************************************************
******************************************************************************

*** the evaluator

******************************************************************************
******************************************************************************
*****************************************************************************/


/* the evaluator is more complex than it might be because it needs to
   maintain lisp pointers into partially evaluated structures as it
   evaluates, but there might be a garbage collection in the meantime
   and those pointers might shift.  therefore it has to make sure that
   all pointers it is refering to are in places that the garbage collector
   has access to and can shift them.

   there are two stacks:
     one is refered to via stack[sp], and this is the evaluation stack
     things are put on the stack via:
        pushstack(item)
	pushstackfresh(item)
     in each case the top of the stack (stack[sp]) then points to a
     cons cell whose car is the item.  for pushstack its cdr is the
     cons cell in stack[sp-1].  for pushstackfresh its cdr is nil.
     new procedure call frames are pushed using pushstackfresh, while
     everything else uses pushstack.  in this way stack[sp] is always
     the complete lexical environment, but previous lexical
     sub-environments can be returned to by decrementing sp appropriately.
     this whole stack is only ever refered to explicitly with exactly
     the form stack[sp].

     the other stack is a "register" stack where gc sensitive items are
     stored.  it is simply an array of values that need to get forwarded
     by the garbage collector and is regstack[regsp].  an item is added
     to the stack with:
        savereg(item)
     and is recovered with
        restorereg(regname)
	getbackreg(index, regname)
     restorereg pops the stack, and does "regname = ...".  getbackreg
     does not pop the stack, but indexes down it by index slots, so
     if index==0 it gets the top thing on the stack, and also does
     a "regname = ...".  there are also ways to refer to the register
     stack and to change contents--these do not pop
        referreg(index)

     the heights of both stacks are saved and restored via
        savestack(place)
	restorestack(place)
     where place is of type sstack.

   eval and loadfile always returns with the stacks restored to where
   they were on entry.  other things (e.g., eval_form) do not.

   there are two sorts of side effects that can happen when some other
   procudures are called.
     there can be things added to the stacks by evaluation
       in this case the stacks must be restored to where they were
       before any further evaluation, or before any retrieval of
       registers

     there can be consing
       in this case any pointers eferred to after that must be retrieved
       via the register stack, and not from C variables directly. if
       there was no evaluation the stacks do not need to be restored first.

   it is possible to know whether either or both of these are possible
   and so the code in the evaluator is written taking what is possible
   into account.

*/

static object throwvalue, throwtag;

#define catchpoints (1000)
static catchpoint catches[catchpoints];

static int currentcatch = -1;

static catchpoint *get_freecatchpoint () {
  catchpoint *ret;
  if (catchpoints<=currentcatch++)
    {printf("\n\nRan out of catch points!!!\n\n");
    exit(-1);}
  ret = &catches[currentcatch];
  ret->free = currentcatch-1;
  return ret;}
	

static void restore_catchpoint (catchpoint *catchp) {
  currentcatch = catchp->free;
  return;}
	 
extern void fell_reploop (void *filename) {
  object form, value;
  catchpoint *maincatch;
	
  if (NULL!=filename)
    loadfile(make_string((char *)filename));
  
 loop:
  currentcatch = -1;
  maincatch = get_freecatchpoint();
  format(tsymbol, prompt, nilobj);
  if (0 != setjmp(maincatch->env))
    {if (throwtag==prompt)
      goto loop;
    else if (throwtag==warnstring)
      {printf("\nExiting FELL...\n");
      return;}
    else {
      format(tsymbol,
	     make_string("~&;;; ERROR: Uncaught throw tag: ~S.  Aborting..."),
	     cons(throwtag,nilobj));
      goto loop;}}
  sp = 0;
  regsp = 0;
  form = readproc(tsymbol, nilobj, nilobj);
  tprintmethod->column = 0;
  symbol(minus)->value = form;
  value = eval(form);
  symbol(plusplusplus)->value = symbol(plusplus)->value;
  symbol(plusplus)->value = symbol(plus)->value;
  symbol(plus)->value = form;
  if (0!=tprintmethod->column) terpri(tsymbol);
  prin1(value, tsymbol);
  symbol(starstarstar)->value = symbol(starstar)->value;
  symbol(starstar)->value = symbol(star)->value;
  symbol(star)->value = value;
  goto loop;
  return;}


static object loadfile (object filename) {
  object str, red;
  sstack savedstack;
  catchpoint *catchplace;
	
  str = fellopen(filename, nilobj);
  savereg(str);
  savestack(savedstack);
  catchplace = get_freecatchpoint();
  if (0 != setjmp(catchplace->env))
    {restore_catchpoint(catchplace);
    restorestack(savedstack);
    restorereg(str);
    fellclose(str);
    simthrow(throwtag,throwvalue);}

  while ((red=readproc(str, nilobj, str),getbackreg(0,str),str!=red))
    eval(red);
    getbackreg(0,str);
  restore_catchpoint(catchplace);
  restorereg(str);
  fellclose(str);
  return tsymbol;}


static object eval_special_warns;
static boolean toplevel_eval;
static boolean toplevel_form;

static object eval (object form) {
  sstack savedstack;
  object ret;
  eval_special_warns = nilobj;
  toplevel_eval = true;
  toplevel_form = true;
  savestack(savedstack);
  ret = eval_form(form);
  restorestack(savedstack);
  return ret;}
	
#define seenspecials 1
#define seenlexicals 2

/* need to deal with
   eval_special_warns
   procedure application onwards
*/


static object eval_form (object form) {
  object found, proced, head, subforms, blocktemp;
  boolean toplevel_formp;
  catchpoint *catchplace;
  int len;
  sstack savedstack;
	
  if (sp>stackbound) error0("Control stack overflow");
  if (regsp>regstackbound) error0("Register stack overflow");

  toplevel_formp = toplevel_form;
  toplevel_form = false;
  background(fell_busy);

  savereg(form);
  found = cons(evalsymbol,form);
  pushstack(found);
  restorereg(form);

  if (notpairp(form))
    {switch(gettype(form)){
    case type_null: case type_fixnum: case type_flonum:
    case type_string: case type_vector:
      return form;
    case type_symbol:

      found = searchvarstack(form, stack[sp]);
      if (nilobj!=found)
	return rest(found);
      else if (nilobj!=is_special_p(form))
	{if (boundp(form))
	  return symbol(form)->value;
	else
	  error1("Unbound special symbol: ~S", form);}
      else if (nilobj!=is_constant_p(form))
	return symbol(form)->value;
      else if (boundp(form))
	{if (nilobj==memq(form, eval_special_warns))
	  {eval_special_warns = cons(form, eval_special_warns);
	  warn1("Free reference to: ~S", form);}
	return symbol(form)->value;}
      else
	error1("Free reference to unbound: ~S", form);
    default: error1("Unevaluable form: ~S", form);}}
  else {
    head = first(form);
    len = 0;
    subforms = form;
    while(pairp(subforms)) {len++; subforms=rest(subforms);}
    if (nilobj!=subforms) error1("Badly terminated eval form: ~S", form);

    if (quote==head) {
      check_specialform(len, 2, 2, form);
      return second(form);}

    else if (prognsymbol==head) {
      check_specialform(len, 1, false, form);
      return eval_form_list(rest(form));}

    else if (ifsymbol==head) {
      check_specialform(len, 3, 4, form);
      savereg(cdr(cdr(form)));
      savestack(savedstack);
      found = eval_form(second(form));
      restorestack(savedstack);
      restorereg(form);
      if (nilobj!=found)
	return eval_form(car(form));
      else
	return eval_form(car(cdr(form)));}

    else if (whensymbol==head) {
      check_specialform(len, 2, false, form);
      savereg(rest(rest(form)));
      savestack(savedstack);
      if (nilobj!=eval_form(second(form))) {
	restorestack(savedstack);
	restorereg(form);
	return eval_form_list(form);}
      else
	return nilobj;}

    else if (unlesssymbol==head) {
      check_specialform(len, 2, false, form);
      savereg(rest(rest(form)));
      savestack(savedstack);
      if (nilobj==eval_form(second(form))) {
	restorestack(savedstack);
	restorereg(form);
	return eval_form_list(form);}
      else
	return nilobj;}

    else if (andsymbol==head) {
      check_specialform(len, 1, false, form);
      if (nullp(rest(form)))
	return tsymbol;
      else
	{form = rest(form);
	while (pairp(rest(form))) {
	  savereg(rest(form));
	  savestack(savedstack);
	  if (nilobj==eval_form(first(form)))
	    return nilobj;
	  restorestack(savedstack);
	  restorereg(form);}
	return eval_form(first(form));}}
    
    else if (orsymbol==head) {
      check_specialform(len, 1, false, form);
      if (nullp(rest(form)))
	return nilobj;
      else
	{form = rest(form);
	while (pairp(rest(form))) {
	  savereg(rest(form));
	  savestack(savedstack);
	  found=eval_form(first(form));
	  if (found!=nilobj)
	    return found;
	  restorestack(savedstack);
	  restorereg(form);}
	return eval_form(first(form));}}

    else if (blocksymbol==head)
      {check_specialform(len, 2, false, form);
      if (!((symbolp(second(form))) || (nilobj==(second(form)))))
	error1("Must have a symbol as a block name: ~S", form);
      savereg(rest(rest(form)));
      blocktemp = cons(blocksymbol,second(form));
      savereg(blocktemp);
      savestack(savedstack);
      catchplace = get_freecatchpoint();
      if (0 != setjmp(catchplace->env))
	{restore_catchpoint(catchplace);
	restorestack(savedstack);
	restorereg(blocktemp);
	if (blocktemp==throwtag)
	  return throwvalue;
	else
	  longjmp(catches[currentcatch].env, 1);}
      getbackreg(0, blocktemp);
      pushstack(blocktemp);		/* this will cons */
      getbackreg(1, form);
      found = eval_form_list(form);
      restore_catchpoint(catchplace);
      return found;}

    else if (returnsymbol==head)
      {check_specialform(len, 1, 2, form);
      eval_return(nilobj, second(form));}

    else if (returnfromsymbol==head)
      {check_specialform(len, 2, 3, form);
      if (!((symbolp(second(form))) || (nilobj==(second(form)))))
	error1("Must have a symbol as a return-from name: ~S", form);
      eval_return(second(form), third(form));}

    else if (catchsymbol==head)
      {check_specialform(len, 2, false, form);
      savereg(form);
      savestack(savedstack);
      found = eval_form(second(form));
      restorestack(savedstack);
      restorereg(form);
      savereg(found);		/* leaves regsp at same place as save */
      catchplace = get_freecatchpoint();
      if (0 != setjmp(catchplace->env))
	{restore_catchpoint(catchplace);
	restorestack(savedstack);
	restorereg(found);
	if (found==throwtag)
	  return throwvalue;
	else
	  longjmp(catches[currentcatch].env, 1);}
      found = eval_form_list(rest(rest(form)));
      restorestack(savedstack);
      restore_catchpoint(catchplace);
      return found;}

    else if (throwsymbol==head)
      {check_specialform(len, 3, 3, form);
      found = eval_form(second(form));
      simthrow(found, eval_form(third(form)));}

    else if (unwindsymbol==head)
      {check_specialform(len, 2, false, form);
      savereg(form);
      savestack(savedstack);
      catchplace = get_freecatchpoint();
      if (0 != setjmp(catchplace->env))
	{restore_catchpoint(catchplace);
	restorestack(savedstack);
	restorereg(form);
	savereg(throwtag);
	savereg(throwvalue);
	savestack(savedstack);
	eval_form_list(rest(rest(form)));
	restorestack(savedstack);
	restorereg(throwvalue);
	restorereg(throwtag);
	simthrow(throwtag,throwvalue);}
      found = eval_form(second(form));
      restore_catchpoint(catchplace);
      restorestack(savedstack);
      restorereg(form);
      eval_form_list(rest(rest(form)));
      return found;}

    else if (setqsymbol==head)
      {check_specialform(len, 3, 3, form);
      savereg(form);
      savestack(savedstack);
      subforms = eval_form(third(form));
      restorestack(savedstack);
      restorereg(form);
      head = second(form);
      if ((!symbolp(head)) || (nilobj!=is_constant_p(head)))
	error1("Can only setq non-constant symbols: ~S", form);
      if (nilobj!=is_special_p(head))
	symbol(head)->value = subforms;
      else
	{found = searchvarstack(head, stack[sp]);
	if (nilobj!=found)
	  cdr(found) = subforms;
	else if (toplevel_eval)
	  {warn1("Setqing non-special ~S -- making it special", head);
	  declare_special(head);
	  symbol(head)->value = subforms;}
	else
	  {warn1("Setqing non-special -- remaining non-special: ~S", form);
	  symbol(head)->value = subforms;}}
      return subforms;}

    else if (letsymbol==head)
      {check_specialform(len, 2, false, form);
      proced = second(form);
      len = check_letbindings(proced); /* not really length at all */
      savereg(proced);
      subforms = (nullp(rest(rest(form))))?nilobj
	:(nullp(rest(rest(rest(form)))))
	?(third(form)):cons(prognsymbol,rest(rest(form))); /* consing */
      if (0==len)
	return eval_form(subforms);
      getbackreg(0,proced);
      savereg(subforms);
      savestack(savedstack);
      found = eval_letbindings(proced);
      restorestack(savedstack);
      savereg(found);
      if (seenlexicals==len) {
	head = cons(letsymbol,nilobj);
	savereg(head);
	/* stack is head, found, subforms, proced */
	while (referreg(1)!=nilobj) {
	  getbackreg(3,proced);
	  /* consing here */
	  pushletbind(symbolp(first(proced))?first(proced)
		      :first(first(proced)),first(referreg(1)),referreg(0));
	  referreg(1) = rest(referreg(1));
	  referreg(3) = rest(referreg(3));}
	pushstack(referreg(0));
	return eval_form(referreg(2));}
      else if (seenspecials==len) {
	head = cons(progvsymbol,nilobj);
	getbackreg(0,found);
	savereg(head);
	/* stack is head, found, subforms, proced */
	while (referreg(1)!=nilobj) {
	  blocktemp = first(referreg(3)); /* proced is source bindings */
	  /* handle symbol only form */
	  blocktemp = (symbolp(blocktemp))?blocktemp:first(blocktemp);
	  pushletbind(blocktemp,symbol(blocktemp)->value,referreg(0));
	  symbol(blocktemp)->value = first(referreg(1));
	  referreg(1) = rest(referreg(1));
	  referreg(3) = rest(referreg(3));}
	pushstack(referreg(0));
	getbackreg(2,subforms);
	regsp -= 4;
	savestack(savedstack);
	catchplace = get_freecatchpoint();
	if (0 == (len = setjmp(catchplace->env)))
	  found = eval_form(subforms);
	restorestack(savedstack);
	restore_catchpoint(catchplace);
	for (subforms=rest(first(stack[sp]));
	     subforms!=nilobj;
	     subforms=rest(subforms))
	  symbol(first(first(subforms)))->value = rest(first(subforms));
	if (0==len)
	  return found;
	else
	  simthrow(throwtag,throwvalue);}
      else error1("FELL does not support mixed lexicals and specials in lets: ~S",
		  form);}

    else if (tagbodysymbol==head) {
      check_specialform(len, 1, false, form);
      savereg(form);
      savestack(savedstack);
      proced = parse_tagbody(rest(form));	/* conses and messes regs */
      restorestack(savedstack);
      restorereg(form);
      pushstack(proced);
      eval_tagbody(rest(form), proced);
      return nilobj;}

    else if (gosymbol==head) {
      check_specialform(len, 2, 2, form);
      found = lookup_gotag(second(form));
      if (nilobj==found)
	error1("Can not find go tag: ~S", second(form));
      else
	simthrow(rest(found),first(found));}

    else if (dotimessymbol==head) {
      check_specialform(len, 2, false, form);
      subforms = second(form);
      check_doform(subforms);
      savereg(subforms);
      savereg(form);
      savestack(savedstack);
      found = eval_form(second(subforms));
      restorestack(savedstack);
      savereg(found);
      savestack(savedstack);
      /* stack is found, form, subforms */
      if (!fixnump(found))
	error2("Bound ~S is not a fixnum in: ~S", found, form);
      proced = parse_tagbody(rest(rest(referreg(1))));
      restorestack(savedstack);
      savereg(proced);
      pushstack(proced);
      head = cons(letsymbol,nilobj);
      savereg(head);
      /* stack is head, proced, found, form, subforms */
      pushletbind(first(referreg(4)),zerobj,referreg(0));
      pushstack(referreg(0));
      savestack(savedstack);
      for (len=0;len<fixnum(referreg(2))->value;len++) {
	eval_tagbody(rest(rest(referreg(3))), referreg(1));
	rest(second(referreg(0)))=make_fixnum(len+1);
	restorestack(savedstack);}
      return eval_form(third(referreg(4)));}

    else if (dolistsymbol==head)
      {check_specialform(len, 2, false, form);
      subforms = second(form);
      check_doform(subforms);
      savereg(subforms);
      savereg(form);
      savestack(savedstack);
      found = eval_form(second(subforms));
      restorestack(savedstack);
      savereg(found);
      savestack(savedstack);
      /* stack is found, form, subforms */
      proced = parse_tagbody(rest(rest(referreg(1))));
      restorestack(savedstack);
      savereg(proced);
      pushstack(proced);
      head = cons(letsymbol,nilobj);
      savereg(head);
      /* stack is head, proced, found, form, subforms */
      pushletbind(first(referreg(4)),nilobj,referreg(0));
      pushstack(referreg(0));
      savestack(savedstack);
      for (;nilobj!=referreg(2);referreg(2)=rest(referreg(2))) {
	if (notpairp(referreg(2)))
	  error1("Iteration list is terminated with ~S", referreg(2));
	rest(second(referreg(0)))=first(referreg(2));
	eval_tagbody(rest(rest(referreg(3))), referreg(1));
	restorestack(savedstack);}
      rest(second(referreg(0)))=nilobj;
      return eval_form(third(referreg(4)));}

    else if (function==head)
      {check_specialform(len, 2, 2, form);
      subforms = second(form);
      if ((symbolp(subforms)) && (procp(symbol(subforms)->procedure)))
	return symbol(subforms)->procedure;
      else if ((pairp(subforms)) && (lambda==first(subforms)))
	{check_specialform(len, 2, false, subforms);
	return defineproc(first(subforms),
			  false,
			  proc_lambda,
			  rest(subforms),
			  stack[sp]);} 
      else
	error1("Not a valid procedure: ~S", form);}
    
    /* now do things that should be macros in some other implementation */
    else if ((defunsymbol==head)||(defmacrosymbol==head))
      {if (!toplevel_formp)
	error1("~S can only appear at toplevel", head);
      check_specialform(len, 3, false, form);
      if (!(symbolp(second(form))))
	error1("Procedure name must be a symbol: ~S", form);
      return defineproc(second(form),
			true,
			(defunsymbol==head)?proc_lambda:proc_macro,
			rest(rest(form)),
			nilobj);}
    else if ((defconstantsymbol==head)
	     ||(defvarsymbol==head)
	     ||(defparametersymbol==head))
      {if (!toplevel_formp)
	error1("~S can only appear at toplevel", head);
      check_specialform(len, 3, 3, form);
      if (!(symbolp(second(form))))
	error1("Constant/var/parameter name must be a symbol: ~S", form);
      savereg(second(form));
      savestack(savedstack);
      found = eval_form(third(form));
      restorestack(savedstack);
      restorereg(subforms);

      if (defconstantsymbol==head)
	{if (nilobj!=is_special_p(subforms))
	  error1("~S is already a special; can't be declared constant",
		 subforms);
	else if (nilobj==is_constant_p(subforms))
	  {declare_constant(subforms);
	  symbol(subforms)->value=found;}
	else if (!bequal(found,symbol(subforms)->value))
	  {warn2("Redefining constant ~S from old value ~S",
		 subforms,
		 symbol(subforms)->value);
	  symbol(subforms)->value=found;}}

      else if (nilobj!=is_constant_p(subforms))
	error1("~S is already a constant; can't be declared special",
	       subforms);

      else if (nilobj==is_special_p(subforms))
	{declare_special(subforms);
	symbol(subforms)->value=found;}

      else if (head==defparametersymbol)
	symbol(subforms)->value=found;
      return second(form);}

    /* it better be a procedure application by this point... */
    else if (symbolp(head)) {
      proced = symbol(head)->procedure;
      savereg(form);
      savereg(proced);
      savestack(savedstack);
      if (procedurep(proced))
	{if ((procp(proced)) || (cprocp(proced))) {
	  found = eval_list(rest(form));
	  restorestack(savedstack);
	  restorereg(proced);
	  return applyproc(proced, found);}
	else if (macroprocp(proced)) {
	  subforms = applyproc(proced, rest(form));
	  restorestack(savedstack);
	  getbackreg(1, form);
	  car(form) = car(subforms); /* snap the links */
	  cdr(form) = cdr(subforms);
	  return eval_form(form);}
	else
	  error1("Can't yet evaluate this sort of procedure call: ~S", form);}
      else
	error1("Unbound symbol in procedure place: ~S", form);}
    else if ((pairp(head))&&(lambda==(first(head)))) {
      check_specialform(felllength(head), 2, false, rest(head));
      savereg(head);
      savestack(savedstack);
      found = eval_list(rest(form));
      restorestack(savedstack);
      restorereg(head);
      return applyproc(defineproc(first(head),
				  false,
				  proc_lambda,
				  rest(head),
				  stack[sp]),
		       found);}
    else 
      {warn1("Can't yet evaluate this form: ~S", form);
      return form;}}}	


static object applyproc (object proced, object arglist) {
  int mnargs, mxargs, len;
  boolean saved_toplevel_eval;
  sstack savedstack;
  object argl;

  if ((primopp(proced)) || (cprocp(proced)))
    {mnargs = minargcount(proced);
    mxargs = maxargcount(proced);
    len = 0;
    argl = arglist;
    while(pairp(argl)) {len++; argl = cdr(argl);}
    if ((len<mnargs)||(mxargs<len))
      error2("Wrong number of arguments to ~S: ~S",
	     procedure(proced)->name,
	     arglist);
    if (cprocp(proced)) return apply_cproc(proced, arglist);
    switch(mnargs){
    case 0:
      if (mxargs==0)
	return ((object(*)())procedure(proced)->proc)();
      else
	return ((object(*)(object))procedure(proced)->proc)(arglist);
    case 1:
      if (mxargs==1)
	return ((object(*)(object))procedure(proced)->proc)(first(arglist));
      else
	return ((object(*)(object, object))procedure(proced)->proc)
	  (first(arglist), rest(arglist));
    case 2:
      if (mxargs==2)
	return ((object(*)(object, object))procedure(proced)->proc)
	  (first(arglist), second(arglist));
      else
	return ((object(*)(object, object, object))procedure(proced)->proc)
	  (first(arglist), second(arglist), rest(rest(arglist)));
    case 3:
      if (mxargs==3)
	return ((object(*)(object, object, object))procedure(proced)->proc)
	  (first(arglist), second(arglist), third(arglist));
      else
	return ((object(*)(object, object, object, object))
		procedure(proced)->proc)
	  (first(arglist),
	   second(arglist),
	   third(arglist),
	   rest(rest(arglist)));
    default:
      error2("Can't yet handle primitive procedure ~S with arguments: ~S",
	     procedure(proced)->name,
	     arglist);}
    }
  else if ((lambdap(proced)) || (macroprocp(proced))) {
    saved_toplevel_eval = toplevel_eval;
    toplevel_eval = false;
    savereg(proced);
    pushstackfresh(bindargs(procedure(proced)->arglist,
			    arglist,
			    proced,
			    arglist));
    restorereg(proced);
    argl = eval_form((object)procedure(proced)->proc);
    toplevel_eval = saved_toplevel_eval;
    return argl;}
  else
    error1("Can't yet apply this sort of procedure: ~S",
	   procedure(proced)->name);}


static object applyspread (object proced, object restargs) {
  object spread;
  savereg(proced);
  spread = applyproc(symbol(l__liststar)->procedure, restargs);
  restorereg(proced);
  return applyproc(proced, spread);}
	

static double stash0, stash1, stash2;

static object apply_cproc (object proced, object restargs) {
  int desc, test;
  void *cresult;
  void *parg0;
  void *parg1;
  void *parg2;
	
  desc = (int)(procedure(proced)->arglist);
  test = 0xff & (desc>>16);
  if (c_void == test)
    cresult = (((void *(*)())(procedure(proced)->proc)))();
  else 
    {parg0 = process_carg(test, first(restargs), &stash0);
    test = 0xff & (desc>>8);
    if (c_void == test)
      cresult = (((void *(*)())(procedure(proced)->proc)))(parg0);
    else
      {parg1 = process_carg(test, second(restargs), &stash1);
      test = 0xff & desc;
      if (c_void == test)
	cresult = (((void *(*)())(procedure(proced)->proc)))(parg0, parg1);
      else
	{parg2 = process_carg(test, third(restargs), &stash2);
	cresult = (((void *(*)())(procedure(proced)->proc)))(parg0,
							     parg1,
							     parg2);}}}
  test = 0xff & (desc>>24);
  if (test == c_int)
    return make_fixnum((int)cresult);
  else if (test == c_pointer)
    return make_cpointer(cresult);
  else if (test == c_boolean)
    return (c_boolean == 0)?nilobj:tsymbol;
  else if ((test == c_void) || (cresult==NULL))
    return nilobj;
  else if (test == c_double)
    return make_flonum(*(double *)cresult);
  else if (test == c_string)
    return make_string((char *)cresult);
  else if (test == c_ltype)
    return (object)cresult;
  else if (test == c_lpointer)
    return (object)cresult;
  else
    error1("Sick and confused in applying a C procedure: ~S", proced);}


static void *process_carg (int typ, object arg, double *stash) {
  if (typ == c_int)
    {if (fixnump(arg))
      return (void *)fixnum(arg)->value;
    else
      error1("C call: supplied argument ~S is not coercible to an int",
	     arg);}
  else if (typ == c_double)
    {if (flonump(arg))
      return (void *)&flonum(arg)->value;
    else if (fixnump(arg))
      {(*stash) = (double)(int)fixnum(arg)->value;
      return (void *)stash;}
    else
      error1("C call: supplied argument ~S is not coercible to a double",
	     arg);}
  else if (typ == c_string)
    {if (stringp(arg))
      return (void *)string(arg)->chars;
    else if (symbolp(arg))
      return (void *)string(symbol(arg)->name)->chars;
    else
      error1("C call: supplied argument ~S is not coercible to a string",
	     arg);}
  else if (typ == c_ltype)
    {if (ltypep(arg))
      return (void *)arg;
    else
      error1("C call: supplied argument ~S is not an allocated type", arg);}
  else if (typ == c_lpointer)
    {if (lpointerp(arg))
      return (void *)arg;
    else
      error1("C call: supplied argument ~S is not a FELL type", arg);}
  else if (typ == c_boolean)
    return (arg==nilobj)?(void *)0:(void *)1;
  else
    {if (cpointerp(arg))
      return cpointer(arg)->point;
    else
      error1("C call: supplied argument ~S is not coercible to a C pointer", 
	     arg);}}	


static object defineproc (object name,
			  boolean named,
			  int flags,
			  object argbody,
			  object env) {
  object body, proced;

  savereg(name);
  savereg(argbody);
  savereg(env);
  check_arglist(first(argbody));
  body = rest(argbody);
  if (pairp(body))
    {if (nullp(rest(body)))
      body = first(body);
    else
      body = cons(prognsymbol, body);}
  savereg(body);
  proced = make_procedure(flags);
  restorereg(body);
  restorereg(env);
  restorereg(argbody);
  restorereg(name);
  if (named) symbol(name)->procedure = proced;
  procedure(proced)->name = name;
  procedure(proced)->arglist = first(argbody);
  procedure(proced)->proc = body;
  procedure(proced)->env = env;
  return (named)?name:proced;}


static object eval_form_list (object forms) {
  sstack savedstack;
  if (forms==nilobj)
    return nilobj;
  savereg(forms);
  savestack(savedstack);
  while(nilobj!=rest(referreg(0)))
    {eval_form(first(referreg(0)));
    restorestack(savedstack);
    referreg(0) = rest(referreg(0));}
  return eval_form(first(referreg(0)));}


static object eval_list (object forms){
  object tmp;
  sstack savedstack;
	
  savereg(forms);
  savereg(nilobj);
  savestack(savedstack);
  while (referreg(1)!=nilobj) {
    tmp = eval_form(first(referreg(1)));
    restorestack(savedstack);
    referreg(0) = cons(tmp, referreg(0));
    referreg(1) = rest(referreg(1));}
  return nreverse(referreg(0));}
	

static void eval_return (object blockname, object form) {
  object stackscan;
	
  for (stackscan=stack[sp];
       (nilobj!=stackscan)&&(framesymbol!=first(first(stackscan)));
       stackscan = rest(stackscan))
    {if ((blocksymbol==first(first(stackscan)))
	 &&(blockname==rest(first(stackscan))))
      simthrow(first(stackscan), eval_form(form));}
  error1("Can't return from block named: ~S", blockname);}


static void simthrow (object tag, object value) {
  if (currentcatch<0)
    error1("Whoa! Trying to throw to ~S, when there are no catches!", tag);
  throwvalue = value;
  throwtag = tag;
  longjmp(catches[currentcatch].env, 1);}
	

static object parse_tagbody (object forms) {
  object sf;
  boolean seentags = false;
  
  for (sf=forms;pairp(sf);sf=rest(sf))
    if (notpairp(first(sf))) seentags=true;
  if (!seentags) return nilobj;
	
  savereg(forms);
  savereg(cons(tagbodysymbol, nilobj));
  for (;pairp(referreg(1));referreg(1)=rest(referreg(1)))
    if (notpairp(first(referreg(1))))
      pushtag(first(referreg(1)), referreg(0));
  pushstack(referreg(0));
  return referreg(0);}


static void eval_tagbody (object forms, object catcher) {
  catchpoint *catchplace;
  sstack savedstack, savedstack1;
  savestack(savedstack1);
  if (nullp(catcher))
    eval_form_list(forms);
  else {
    catchplace = get_freecatchpoint();
    savereg(catcher);
    savereg(forms);
    savereg(forms);
    savereg(forms);
    savestack(savedstack);
    /* stack is sf, useforms, forms, catcher */
    loop:
    if (0 != setjmp(catchplace->env)) {
      if (referreg(3)==throwtag) {
	restorestack(savedstack);
	referreg(1) = memq(throwvalue, referreg(2));
	goto loop;}
      else {
	restore_catchpoint(catchplace);
	longjmp(catches[currentcatch].env, 1);}}
    for (referreg(0)=referreg(1);
	 pairp(referreg(0));
	 referreg(0)=rest(referreg(0)))
      if (pairp(first(referreg(0)))) {
	eval_form(first(referreg(0)));
	restorestack(savedstack);}
    restore_catchpoint(catchplace);}
  restorestack(savedstack1);
  return;}


static void check_specialform (int len, int lo, int hi, object form) {
  if ((len<lo)
      ||((hi!=false)&(len>hi)))
    error1("Wrong number of subforms in: ~S", form);
  return;}


static void check_doform (object subf) {
  if (! ((pairp(subf))
	 &&(pairp(rest(subf)))
	 &&(nullp(rest(rest(rest(subf)))))
	 &&(symbolp(first(subf)))))
    error1("Malformed variable and stepper in do form: ~S", subf);
  return;}
	
	
static void check_arglist (object arglist) {
  object argl, argname, optionform;
  int firstchar;
  boolean hasoption;
  boolean seenopt = false;
  boolean seenrest = false;
  boolean seenkey = false;
  
  argl = arglist;
  while (pairp(argl))
    {argname = first(argl);
    if (symbolp(argname))
      hasoption = false;
    else
      {if (!((pairp(argname))
	     &&(symbolp(first(argname)))
	     &&(pairp(rest(argname)))
	     &&(nullp(rest(rest(argname))))))
	error2("Malformed (optional?) argument: ~S in ~S", argname, arglist);
      optionform = second(argname);
      argname = first(argname);
      hasoption = true;}
    argl = rest(argl);
    if (hasoption && ((!seenopt)||seenrest))
      error2("Optional form for ~S occurs in illegal place: ~S",
	     argname,
	     arglist);
    firstchar = string(symbol(argname)->name)->chars[0];
    if (nilobj!=is_special_p(argname))
      error2("Argument can not be a special: ~S in ~S", argname, arglist);
    if (nilobj!=is_constant_p(argname))
      error2("Argument can not be a constant: ~S in ~S", argname, arglist);
    if (':'==firstchar)
      error2("Argument can not be a keyword: ~S in ~S", argname, arglist);
    if ('&'==firstchar)
      {if (argname==andrestsymbol)
	{if (seenkey)
	  error1("Can't have &rest follow &key: ~S", arglist);
	if (seenrest)
	  error1("Can't have two &rest's: ~S", arglist);
	if (!(pairp(argl)) && (nullp(rest(argl))) && (symbolp(first(argl))))
	  error1("Malformed rest arg specification: ~S", arglist);
	seenrest = true;}
      else if (argname==andoptsymbol)
	{if (seenrest || seenkey || seenopt)
	  error1("An &optional occurs after a keyword in argument list: ~S",
		 arglist);
	seenopt = true;}
      else if (argname==andkeysymbol)
	{if (seenrest || seenkey || seenopt)
	  error1("An &key occurs after a keyword in argument list: ~S",
		 arglist);
	seenkey = true;}
      else error2("Unknown keyword ~S in argument list: ~S",
		  argname,
		  arglist);}
    }
  if (argl!=nilobj)
    error1("Improper termination of arglist: ~S", arglist);
  return;}
	

static object bindargs (object formals, object actuals, object proced, object arglist) {
  object formalarg, result;
  sstack savedstack, savedstack1;
  boolean scanopt = false;

  savestack(savedstack1);
  savereg(arglist);
  savereg(proced);
  savereg(actuals);
  savereg(formals);
  result = makeframe(procedure(proced)->name, procedure(proced)->env);
  savereg(result);
  /* stack is result, formals, actuals, proced, arglist */
  while(nilobj!=referreg(1)) {
    formalarg = first(referreg(1));
    if (formalarg==andrestsymbol) {
      pushframe(second(referreg(1)), referreg(2), referreg(0));
      referreg(2) = nilobj;
      referreg(1) = nilobj;}
    else if (formalarg==andoptsymbol)
      {scanopt = true;
      referreg(1) = rest(referreg(1));}
    else if (scanopt&&(nilobj==referreg(2))) {
      if (symbolp(formalarg))
	pushframe(formalarg,nilobj,referreg(0));
      else {
	savestack(savedstack);
	pushstack(referreg(0));
	formalarg = eval_form(second(formalarg));
	restorestack(savedstack);
	pushframe(first(first(referreg(1))),
		  formalarg,
		  referreg(0));}
      referreg(1) = rest(referreg(1));}
    else
      {if (scanopt && (!(symbolp(formalarg))))
	formalarg = first(formalarg);
      if (nilobj==referreg(2))
	error2("Too few arguments supplied to ~S: ~S",
	       procedure(referreg(3))->name,
	       referreg(4));
      pushframe(formalarg,first(referreg(2)),referreg(0));
      referreg(2) = rest(referreg(2));
      referreg(1) = rest(referreg(1));}}
  if (nilobj!=referreg(2))
    error2("Too many arguments supplied to ~S: ~S",
	   procedure(referreg(3))->name,
	   referreg(4));
  restorereg(result);
  restorestack(savedstack1);
  return result;}


static object searchvarstack(object var, object place) {
  object stackscan, keysym, searchplace, frame;

  if (place==nilobj) return nilobj;
  frame = nilobj;
  for(stackscan=place;stackscan!=nilobj;stackscan=cdr(stackscan)){
    keysym = first(first(stackscan));
    searchplace = nilobj;
    if (keysym==framesymbol) {	/* must be the end of the env... */
      frame = first(stackscan);
      searchplace = cdr(cdr(first(stackscan)));}
    else if (keysym==letsymbol)
      searchplace = cdr(first(stackscan));
    while(searchplace!=nilobj) {
      if (first(first(searchplace))==var)
	return first(searchplace);
      searchplace = rest(searchplace);}}
  if ((frame!=nilobj) && (nilobj!=lexenv(frame)))
    return searchvarstack(var, lexenv(frame));
  return nilobj;}


static int check_letbindings (object pairforms) {
  object pfs, varname, binding;
  int result = 0;
	
  pfs = pairforms;
  while (pfs!=nilobj)
    {if (notpairp(pfs))
      error1("Invalid termination of let bindings: ~S", pairforms);
    binding = first(pfs);
    if (symbolp(binding)
	||((pairp(binding))
	   &&(pairp(rest(binding)))
	   &&(nullp(rest(rest(binding))))))
      {varname = symbolp(binding)?binding:first(binding);
      if (!symbolp(varname))
	error2("Non symbol ~S in variable name place in let bindings ~S",
	       varname,
	       pairforms);
      if (nilobj!=is_constant_p(varname))
	error2("Can't rebind constant ~S in let bindings ~S",
	       varname,
	       pairforms);
      result |= ((nilobj!=is_special_p(varname))?seenspecials:seenlexicals);}
    else error2("Badly formed let binding ~S in ~S", binding, pairforms);
    pfs = rest(pfs);}
  return result;}


static object eval_letbindings (object pairforms) {
  object result, temp;
  sstack savedstack;
	
  savereg(pairforms);
  savereg(nilobj);
  savestack(savedstack);
  /* stack is result, pairforms */
  while (referreg(1)!=nilobj) {
    if (pairp(first(referreg(1)))) {
      temp = eval_form(second(first(referreg(1))));
      restorestack(savedstack);
      referreg(0) = cons(temp, referreg(0));}
    else
      referreg(0) = cons(nilobj,referreg(0));
    referreg(1) = rest(referreg(1));}
  return nreverse(referreg(0));}


static object lookup_gotag (object tag) {
  object stackscan, keysym, searchplace;

  for (stackscan=stack[sp]; stackscan!=nilobj; stackscan=cdr(stackscan))
    {keysym = first(first(stackscan));
    if (keysym==framesymbol)
      break;
    else if (keysym==tagbodysymbol)
      for (searchplace=rest(first(stackscan));
	   searchplace!=nilobj;
	   searchplace=cdr(searchplace))
	if (first(first(searchplace))==tag)
	  return first(searchplace);}
  return nilobj;}


static object macroexpand1 (object form) {
  object head, proc;
  if (notpairp(form))
    return form;
  head = first(form);
  if (!symbolp(head))
    return form;
  proc = symbol(head)->procedure;
  if (!macroprocp(proc))
    return form;
  return applyproc(proc, rest(form));}
    

/*****************************************************************************
******************************************************************************
******************************************************************************

*** the inspector

******************************************************************************
******************************************************************************
*****************************************************************************/

static object isaved_plevel, isaved_plength, isaved_pbase;

static object inspect (object item) {
  isaved_plevel = symbol(printlevel)->value;
  isaved_plength = symbol(printlength)->value;
  isaved_pbase = symbol(printbase)->value;
  symbol(printlevel)->value = symbol(inspectprintlevel)->value;
  symbol(printlength)->value = symbol(inspectprintlength)->value;
  symbol(printbase)->value = symbol(inspectprintbase)->value;
  inspect_item(item);
  exit_inspector();
  simthrow(prompt,nilobj);}


static void exit_inspector () {
  symbol(printlevel)->value = isaved_plevel;
  symbol(printlength)->value = isaved_plength;
  symbol(printbase)->value = isaved_pbase;
  return;}


#define inspect_noslots -10

static void inspect_item (object item) {
  int mn, mx, len, i, index;
  object token, conscell, scan;
  savereg(item);
  conscell = cons(item, nilobj);
  restorereg(item);
 redisplay:
  mn = inspect_noslots;
  mx = inspect_noslots;
  format(tsymbol, inspectstring0, conscell);
  switch(gettype(item)) {
  case type_null: case type_marker: case type_cpointer: break;
  case type_fixnum: printf("   -- #x%8.8x", fixnum(item)->value); break;
  case type_flonum:
    printf("   -- #x%8.8x %8.8x", fixnum(item+1)->value, fixnum(item)->value);
    break;
  case type_string: printf("   -- (%d)", string(item)->length); break;
  case type_symbol:
    mn = 0;
    mx = 2;
    printf("\n\n 0: <name>");
    car(conscell) = symbol(item)->name;
    format(tsymbol, inspectstring4, conscell);
    printf("\n 1: <val> ");
    car(conscell) = symbol(item)->value;
    format(tsymbol, inspectstring4, conscell);
    printf("\n 2: <proc>");
    car(conscell) = symbol(item)->procedure;
    format(tsymbol, inspectstring4, conscell);
    break;
  case type_procedure:
    mn = 0;
    mx = 3;
    printf("   -- #x%8.8x\n\n 0: <name>", procedure(item)->info);
    car(conscell) = procedure(item)->name;
    format(tsymbol, inspectstring4, conscell);
    printf("\n 1: <code>");
    car(conscell) = procedure(item)->proc;
    format(tsymbol, inspectstring4, conscell);
    printf("\n 2: <argl>");
    car(conscell) = procedure(item)->arglist;
    format(tsymbol, inspectstring4, conscell);
    printf("\n 3: <env> ");
    car(conscell) = procedure(item)->env;
    format(tsymbol, inspectstring4, conscell);
    break;
  case type_vector:
    len = vector(item)->length;
    printf("   -- (%d)\n", len);
    mn = 0;
    mx = len-1;
    for(i=0;i<len;i++) {
      printf("\n%2d:", i);
      car(conscell) = vector(item)->arr[i];
      format(tsymbol, inspectstring4, conscell);}
    break;
  case type_cons:
    mn = 0;
    mx = -1;
    printf("\n");
    for(scan=item;pairp(scan);scan=pair(scan)->cdr) {
      printf("\n%2d:", ++mx);
      car(conscell) = pair(scan)->car;
      format(tsymbol, inspectstring4, conscell);}
    if (nilobj!=scan) {
      mn = -1;
      printf("\n-1:");
      car(conscell) = scan;
      format(tsymbol, inspectstring4, conscell);}
    break;
  case type_structure:
  }

 readtoken:
  format(tsymbol, inspectstring1, nilobj);
  savereg(conscell);
  savereg(item);
  token = readproc(tsymbol, nilobj, nilobj);
  restorereg(item);
  restorereg(conscell);
  tprintmethod->column = 0;
  if (inspecth==token) {
    format(tsymbol, inspectstring3, nilobj);
    if (mn!=inspect_noslots)
      format(tsymbol, inspectstring5, list2(make_fixnum(mn), make_fixnum(mx)));
    goto readtoken;}
  else if (inspectq==token) {
    exit_inspector();
    simthrow(prompt,nilobj);}
  else if (inspectr==token) goto redisplay;
  else if (inspectu==token) return;
  else if (inspectsb==token) {
    savereg(conscell);
    savereg(item);
    token = readproc(tsymbol, nilobj, nilobj);
    restorereg(item);
    restorereg(conscell);
    if (debug_valid_p(token, false, 2))
      {symbol(printbase)->value = token;
      symbol(inspectprintbase)->value = token;}
    goto readtoken;}
  else if (inspectsln==token) {
    savereg(conscell);
    savereg(item);
    token = readproc(tsymbol, nilobj, nilobj);
    restorereg(item);
    restorereg(conscell);
    if (debug_valid_p(token, true, 1)) {
      symbol(printlength)->value = token;
      symbol(inspectprintlength)->value = token;}
    goto readtoken;}
  else if (inspectslv==token) {
    savereg(conscell);
    savereg(item);
    token = readproc(tsymbol, nilobj, nilobj);
    restorereg(item);
    restorereg(conscell);
    if (debug_valid_p(token, true, 1)) {
      symbol(printlevel)->value = token;
      symbol(inspectprintlevel)->value = token;}
    goto readtoken;}
  else if (fixnump(token)) {
    if (mn==inspect_noslots) {
      format(tsymbol, inspectstring6, nilobj);
      goto readtoken;}
    else if ((mn>fixnum(token)->value)
	     ||(mx<fixnum(token)->value)) {
      format(tsymbol, inspectstring7, nilobj);
      goto readtoken;}
    else {
      index = fixnum(token)->value;
      switch (gettype(item)) {
      case type_vector:
	inspect_item(vector(item)->arr[index]);
	goto redisplay;
      case type_procedure:
	inspect_item(index==0?procedure(item)->name:
		     index==1?procedure(item)->proc:
		     index==2?procedure(item)->arglist:procedure(item)->env);
	goto redisplay;
      case type_symbol:
	inspect_item(index==0?symbol(item)->name:
		     index==1?symbol(item)->value:symbol(item)->procedure);
	goto redisplay;
      case type_cons:
	i = index;
	for(scan=item;pairp(scan);scan=pair(scan)->cdr)
	  if (0==i--) {
	    inspect_item(pair(scan)->car);
	    goto redisplay;}
	inspect_item(scan);
	goto redisplay;}}}
  else {
    format(tsymbol, inspectstring2, nilobj);
    goto readtoken;}}
